#include "disassembler.h"
#include "wtruntime/byteorder.h"
#include "wtruntime/stdlib.h"
#include "labelmap.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NUM_INSTR 256
#define NUM_MICRO ((WTR_NUM_MICROPROGRAMS + 15) & -16)
#define INDENT    "  "

#define __WTR_FILL_MICROPROGS(lowercase,UPPERCASE) \
	do \
	{ \
		disasm->microprognames[WTMP_##UPPERCASE] = #lowercase; \
	} \
	while(0);

#define ADDR_ABSOLUTE 0
#define ADDR_RELATIVE 1

#define ADD_MBZ(_opcode_,_mnemonic_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_mbz; \
		disasm->decoder[(_opcode_)].width    = 1; \
		disasm->decoder[(_opcode_)].custom   = 0; \
	} \
	while(0)

#define ADD_UIMM(_opcode_,_mnemonic_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_uimm; \
		disasm->decoder[(_opcode_)].width    = 1; \
		disasm->decoder[(_opcode_)].custom   = 0; \
	} \
	while(0)

#define ADD_ULIMM(_opcode_,_mnemonic_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_ulimm; \
		disasm->decoder[(_opcode_)].width    = 2; \
		disasm->decoder[(_opcode_)].custom   = 0; \
	} \
	while(0)

#define ADD_ADDR(_opcode_,_mnemonic_,_mode_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_addr; \
		disasm->decoder[(_opcode_)].width    = 1; \
		disasm->decoder[(_opcode_)].custom   = (_mode_); \
	} \
	while(0)

#define ADD_LADDR(_opcode_,_mnemonic_,_mode_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_laddr; \
		disasm->decoder[(_opcode_)].width    = 2; \
		disasm->decoder[(_opcode_)].custom   = (_mode_); \
	} \
	while(0)

#define ADD_EXEC(_opcode_,_mnemonic_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_exec; \
		disasm->decoder[(_opcode_)].width    = 1; \
		disasm->decoder[(_opcode_)].custom   = 0; \
	} \
	while(0)

#define ADD_LEXEC(_opcode_,_mnemonic_) \
	do \
	{ \
		disasm->decoder[(_opcode_)].mnemonic = (_mnemonic_); \
		disasm->decoder[(_opcode_)].decode   = disasm_decode_lexec; \
		disasm->decoder[(_opcode_)].width    = 2; \
		disasm->decoder[(_opcode_)].custom   = 0; \
	} \
	while(0)

struct InstructionDecoder;

struct InstructionDecoder
{
	const char* mnemonic;
	wt_bool   (*decode)(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels);
	size_t      width;  /* instruction width in 16-bit words */
	size_t      custom; /* custom parameter to the print function */
};

struct Disassembler
{
	struct InstructionDecoder decoder[NUM_INSTR];
	const char*               microprognames[NUM_MICRO];
	size_t                    num_microprogs;
	wt_bool                   a_version_match;
	wt_bo                     a_byteorder;
};

static wt_u16 disasm_read(struct Disassembler* const disasm, const wt_u16 w)
{
	return wtr_botoh16(w, disasm->a_byteorder);
}

static wt_bool disasm_decode_mbz(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	printf(INDENT INDENT "%08X: %04X       %s\n", addr, instr1, id->mnemonic);
	return WT_TRUE;
}

static wt_bool disasm_decode_uimm(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u32 opdata = WT_TOMB_GETOPDATA(instr1);
	printf(INDENT INDENT "%08X: %04X       %s %u ; 0x%X\n", addr, instr1, id->mnemonic, opdata, opdata);
	return WT_TRUE;
}

static wt_bool disasm_decode_ulimm(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u16 instr2 = disasm_read(disasm, code[1]);
	wt_u32 opdata = (WT_TOMB_GETOPDATA(instr1) << WT_TOMB_INSTRBITS) | instr2;
	printf(INDENT INDENT "%08X: %04X %04X  %s %u ; 0x%X\n", addr, instr1, instr2, id->mnemonic, opdata, opdata);
	return WT_TRUE;
}

static wt_bool disasm_decode_addr(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u32 opdata = WT_TOMB_GETOPDATA(instr1);
	wt_u32 base   = (id->custom==ADDR_RELATIVE) ? addr : 0;
	wt_u32 target = base + opdata*sizeof(*code);
	const char* sign = "";
	const char* lab = NULL;

	if((id->custom==ADDR_RELATIVE) && (opdata & WT_TOMB_HIBITIMM))
	{
		opdata |= ~WT_TOMB_OPDATAMASK;
		target = (base + opdata*sizeof(*code));
		sign = "-";
		opdata = (~opdata) + 1;
	}

	lab = labels ? labelmap_find(labels, target) : NULL;

	printf(INDENT INDENT "%08X: %04X       %s %s0x%X ; <%s%s0x%X>\n", addr, instr1, id->mnemonic, sign, opdata,
		lab ? lab : "", lab ? " @" : "", target);

	return WT_TRUE;
}

static wt_bool disasm_decode_laddr(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u16 instr2 = disasm_read(disasm, code[1]);
	wt_u32 opdata = (WT_TOMB_GETOPDATA(instr1) << WT_TOMB_INSTRBITS) | instr2;
	wt_u32 base   = (id->custom==ADDR_RELATIVE) ? addr : 0;
	wt_u32 target = base + opdata*sizeof(*code);
	const char* sign = "";
	const char* lab = NULL;

	if((id->custom==ADDR_RELATIVE) && (opdata & WT_TOMB_HIBITLONGIMM))
	{
		opdata |= ~WT_TOMB_MASKLONGIMM;
		target = (base + opdata*sizeof(*code));
		sign = "-";
		opdata = (~opdata) + 1;
	}

	lab = labels ? labelmap_find(labels, target) : NULL;

	printf(INDENT INDENT "%08X: %04X %04X  %s %s0x%X ; <%s%s0x%X>\n", addr, instr1, instr2, id->mnemonic, sign, opdata,
		lab ? lab : "", lab ? " @" : "", target);

	return WT_TRUE;
}

static wt_bool disasm_decode_exec(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u32 opdata = WT_TOMB_GETOPDATA(instr1);
	if(disasm->a_version_match)
	{
		if(opdata>=disasm->num_microprogs) { return WT_FALSE; }
		printf(INDENT INDENT "%08X: %04X       %s 0x%X ; <%s>\n", addr, instr1, id->mnemonic, opdata, disasm->microprognames[opdata]);
	}
	else
	{
		printf(INDENT INDENT "%08X: %04X       %s 0x%X\n", addr, instr1, id->mnemonic, opdata);
	}
	return WT_TRUE;
}

static wt_bool disasm_decode_lexec(struct InstructionDecoder* const id, struct Disassembler* const disasm, wt_u32 addr, const wt_u16* code, struct labelmap* const labels)
{
	wt_u16 instr1 = disasm_read(disasm, code[0]);
	wt_u16 instr2 = disasm_read(disasm, code[1]);
	wt_u32 opdata = (WT_TOMB_GETOPDATA(instr1) << WT_TOMB_INSTRBITS) | instr2;
	if(opdata>=disasm->num_microprogs) { return WT_FALSE; }
	printf(INDENT INDENT "%08X: %04X %04X  %s 0x%X ; <%s>\n", addr, instr1, instr2, id->mnemonic, opdata, disasm->microprognames[opdata]);
	return WT_TRUE;
}

struct Disassembler* Disassembler_Create(void)
{
	struct Disassembler* disasm;

	if(NULL!=(disasm=(struct Disassembler*)malloc(sizeof(struct Disassembler))))
	{
		memset(disasm, 0, sizeof(struct Disassembler));

		{
			ADD_MBZ(WT_TOMB_OP_SCMPB, "scmpb");
			ADD_MBZ(WT_TOMB_OP_SCMPH, "scmph");
			ADD_UIMM(WT_TOMB_OP_MSKBTM, "mskbtm");
			ADD_UIMM(WT_TOMB_OP_SGNXD, "sgnxd");
			ADD_UIMM(WT_TOMB_OP_ADDB_I, "addb.i");
			ADD_MBZ(WT_TOMB_OP_ADDB_S, "addb.s");
			ADD_UIMM(WT_TOMB_OP_SUBB_I, "subb.i");
			ADD_MBZ(WT_TOMB_OP_SUBB_S, "subb.s");
			ADD_MBZ(WT_TOMB_OP_MULB, "mulb");
			ADD_MBZ(WT_TOMB_OP_SMULB, "smulb");
			ADD_MBZ(WT_TOMB_OP_DIVB, "divb");
			ADD_MBZ(WT_TOMB_OP_SDIVB, "sdivb");
			ADD_MBZ(WT_TOMB_OP_MODB, "modb");
			ADD_MBZ(WT_TOMB_OP_SMODB, "smodb");
			ADD_UIMM(WT_TOMB_OP_SHLB_I, "shlb.i");
			ADD_MBZ(WT_TOMB_OP_SHLB_S, "shlb.s");
			ADD_MBZ(WT_TOMB_OP_RORB, "rorb");
			ADD_MBZ(WT_TOMB_OP_BNOTB, "bnotb");
			ADD_UIMM(WT_TOMB_OP_ADDH_I, "addh.i");
			ADD_MBZ(WT_TOMB_OP_ADDH_S, "addh.s");
			ADD_UIMM(WT_TOMB_OP_SUBH_I, "subh.i");
			ADD_MBZ(WT_TOMB_OP_SUBH_S, "subh.s");
			ADD_MBZ(WT_TOMB_OP_MULH, "mulh");
			ADD_MBZ(WT_TOMB_OP_SMULH, "smulh");
			ADD_MBZ(WT_TOMB_OP_DIVH, "divh");
			ADD_MBZ(WT_TOMB_OP_SDIVH, "sdivh");
			ADD_MBZ(WT_TOMB_OP_MODH, "modh");
			ADD_MBZ(WT_TOMB_OP_SMODH, "smodh");
			ADD_UIMM(WT_TOMB_OP_SHLH_I, "shlh.i");
			ADD_MBZ(WT_TOMB_OP_SHLH_S, "shlh.s");
			ADD_MBZ(WT_TOMB_OP_RORH, "rorh");
			ADD_MBZ(WT_TOMB_OP_BNOTH, "bnoth");

			/***********************************************************/

			ADD_MBZ(WT_TOMB_OP_TTD, "ttd");

			/***********************************************************/

			ADD_MBZ(WT_TOMB_OP_OSDR, "osdr");

			/***********************************************************/

			ADD_MBZ(WT_TOMB_OP_NOP, "nop");

			ADD_UIMM(WT_TOMB_OP_LD_I, "ld.i");
			ADD_ULIMM(WT_TOMB_OP_LD_L, "ld.l");
			ADD_MBZ(WT_TOMB_OP_LD_S, "ld.s");

			ADD_UIMM(WT_TOMB_OP_ST_I, "st.i");
			ADD_ULIMM(WT_TOMB_OP_ST_L, "st.l");
			ADD_MBZ(WT_TOMB_OP_ST_S, "st.s");

			ADD_UIMM(WT_TOMB_OP_IDRD_I, "idrd.i");
			ADD_ULIMM(WT_TOMB_OP_IDRD_L, "idrd.l");
			ADD_MBZ(WT_TOMB_OP_IDRD_S, "idrd.s");

			ADD_UIMM(WT_TOMB_OP_PBA_I, "pba.i");
			ADD_ULIMM(WT_TOMB_OP_PBA_L, "pba.l");
			ADD_MBZ(WT_TOMB_OP_PBA_S, "pba.s");

			ADD_UIMM(WT_TOMB_OP_PVT_I, "pvt.i");
			ADD_ULIMM(WT_TOMB_OP_PVT_L, "pvt.l");
			ADD_MBZ(WT_TOMB_OP_PVT_S, "pvt.s");
			ADD_MBZ(WT_TOMB_OP_PPC, "ppc");
			ADD_UIMM(WT_TOMB_OP_DUP_I, "dup.i");
			ADD_ULIMM(WT_TOMB_OP_DUP_L, "dup.l");
			ADD_UIMM(WT_TOMB_OP_POP_I, "pop.i");
			ADD_ULIMM(WT_TOMB_OP_POP_L, "pop.l");
			ADD_UIMM(WT_TOMB_OP_PWZ_I, "pwz.i");
			ADD_ULIMM(WT_TOMB_OP_PWZ_L, "pwz.l");
			ADD_UIMM(WT_TOMB_OP_PWO_I, "pwo.i");
			ADD_ULIMM(WT_TOMB_OP_PWO_L, "pwo.l");
			ADD_UIMM(WT_TOMB_OP_PWZL_I, "pwzl.i");
			ADD_ULIMM(WT_TOMB_OP_PWZL_L, "pwzl.l");
			ADD_UIMM(WT_TOMB_OP_PWOL_I, "pwol.i");
			ADD_ULIMM(WT_TOMB_OP_PWOL_L, "pwol.l");
			ADD_UIMM(WT_TOMB_OP_DVTA_I, "dvta.i");
			ADD_ULIMM(WT_TOMB_OP_DVTA_L, "dvta.l");
			ADD_MBZ(WT_TOMB_OP_DVTA_S, "dvta.s");
			ADD_MBZ(WT_TOMB_OP_DVTF, "dvtf");
			ADD_UIMM(WT_TOMB_OP_OSA_I, "osa.i");
			ADD_ULIMM(WT_TOMB_OP_OSA_L, "osa.l");
			ADD_MBZ(WT_TOMB_OP_OSA_S, "osa.s");
			ADD_MBZ(WT_TOMB_OP_OSF, "osf");
			ADD_MBZ(WT_TOMB_OP_OSPU, "ospu");
			ADD_MBZ(WT_TOMB_OP_OSPO, "ospo");
			ADD_MBZ(WT_TOMB_OP_OSSP, "ossp");
			ADD_MBZ(WT_TOMB_OP_TSZ, "tsz");
			ADD_MBZ(WT_TOMB_OP_TPU, "tpu");
			ADD_MBZ(WT_TOMB_OP_TPO, "tpo");
			ADD_MBZ(WT_TOMB_OP_TDR, "tdr");
			ADD_MBZ(WT_TOMB_OP_TCL, "tcl");
			ADD_UIMM(WT_TOMB_OP_PDSP_I, "pdsp.i");
			ADD_ULIMM(WT_TOMB_OP_PDSP_L, "pdsp.l");
			ADD_UIMM(WT_TOMB_OP_VAS_I, "vas.i");
			ADD_ULIMM(WT_TOMB_OP_VAS_L, "vas.l");
			ADD_ADDR(WT_TOMB_OP_ENTER_I, "enter.i", ADDR_ABSOLUTE);
			ADD_LADDR(WT_TOMB_OP_ENTER_L, "enter.l", ADDR_ABSOLUTE);
			ADD_MBZ(WT_TOMB_OP_ENTER_S, "enter.s");
			ADD_MBZ(WT_TOMB_OP_LEAVE, "leave");
			ADD_ADDR(WT_TOMB_OP_JMP_I, "jmp.i", ADDR_RELATIVE);
			ADD_LADDR(WT_TOMB_OP_JMP_L, "jmp.l", ADDR_RELATIVE);
			ADD_MBZ(WT_TOMB_OP_JMP_S, "jmp.s");

			ADD_ADDR(WT_TOMB_OP_JT_I, "jt.i", ADDR_RELATIVE);
			ADD_LADDR(WT_TOMB_OP_JT_L, "jt.l", ADDR_RELATIVE);
			ADD_MBZ(WT_TOMB_OP_JT_S, "jt.s");

			ADD_ADDR(WT_TOMB_OP_JF_I, "jf.i", ADDR_RELATIVE);
			ADD_LADDR(WT_TOMB_OP_JF_L, "jf.l", ADDR_RELATIVE);
			ADD_MBZ(WT_TOMB_OP_JF_S, "jf.s");

			ADD_ADDR(WT_TOMB_OP_JA_I, "ja.i", ADDR_ABSOLUTE);
			ADD_LADDR(WT_TOMB_OP_JA_L, "ja.l", ADDR_ABSOLUTE);
			ADD_MBZ(WT_TOMB_OP_JA_S, "ja.s");

			ADD_EXEC(WT_TOMB_OP_EXEC_I, "exec.i");
			ADD_LEXEC(WT_TOMB_OP_EXEC_L, "exec.l");
			ADD_MBZ(WT_TOMB_OP_EXEC_S, "exec.s");

			ADD_UIMM(WT_TOMB_OP_DLE_I, "dle.i");
			ADD_ULIMM(WT_TOMB_OP_DLE_L, "dle.l");
			ADD_MBZ(WT_TOMB_OP_DLE_S, "dle.s");
			ADD_MBZ(WT_TOMB_OP_CFP, "cfp");
			ADD_ADDR(WT_TOMB_OP_MKFP_I, "mkfp.i", ADDR_ABSOLUTE);
			ADD_LADDR(WT_TOMB_OP_MKFP_L, "mkfp.l", ADDR_ABSOLUTE);
			ADD_MBZ(WT_TOMB_OP_MKFP_S, "mkfp.s");

			ADD_MBZ(WT_TOMB_OP_LDFP, "ldfp");
			ADD_MBZ(WT_TOMB_OP_STFP, "stfp");
			ADD_UIMM(WT_TOMB_OP_FPS_I, "fps.i");
			ADD_ULIMM(WT_TOMB_OP_FPS_L, "fps.l");
			ADD_MBZ(WT_TOMB_OP_STRN, "strn");
			ADD_MBZ(WT_TOMB_OP_STRC, "strc");
			ADD_MBZ(WT_TOMB_OP_STRL, "strl");
			ADD_MBZ(WT_TOMB_OP_STRD, "strd");
			ADD_UIMM(WT_TOMB_OP_STRCR_I, "strcr.i");
			ADD_ULIMM(WT_TOMB_OP_STRCR_L, "strcr.l");
			ADD_MBZ(WT_TOMB_OP_STRCR_S, "strcr.s");
			ADD_UIMM(WT_TOMB_OP_SEQN_I, "seqn.i");
			ADD_ULIMM(WT_TOMB_OP_SEQN_L, "seqn.l");
			ADD_MBZ(WT_TOMB_OP_SEQN_S, "seqn.s");
			ADD_MBZ(WT_TOMB_OP_SEQD, "seqd");
			ADD_UIMM(WT_TOMB_OP_SEQIR_I, "seqir.i");
			ADD_ULIMM(WT_TOMB_OP_SEQIR_L, "seqir.l");
			ADD_MBZ(WT_TOMB_OP_SEQIR_S, "seqir.s");
			ADD_MBZ(WT_TOMB_OP_SEQL, "seql");
			ADD_UIMM(WT_TOMB_OP_RECN_I, "recn.i");
			ADD_ULIMM(WT_TOMB_OP_RECN_L, "recn.l");
			ADD_MBZ(WT_TOMB_OP_RECN_S, "recn.s");
			ADD_MBZ(WT_TOMB_OP_RECD, "recd");
			ADD_UIMM(WT_TOMB_OP_RECFR_I, "recfr.i");
			ADD_ULIMM(WT_TOMB_OP_RECFR_L, "recfr.l");
			ADD_MBZ(WT_TOMB_OP_RECFR_S, "recfr.s");

			ADD_MBZ(WT_TOMB_OP_ERAISE, "eraise");
			ADD_ADDR(WT_TOMB_OP_EXHAN_I, "exhan.i", ADDR_RELATIVE);
			ADD_LADDR(WT_TOMB_OP_EXHAN_L, "exhan.l", ADDR_RELATIVE);
			ADD_MBZ(WT_TOMB_OP_EXHAN_S, "exhan.s");
			ADD_MBZ(WT_TOMB_OP_EHCL, "ehcl");
			ADD_UIMM(WT_TOMB_OP_ADD_I, "add.i");
			ADD_MBZ(WT_TOMB_OP_ADD_S, "add.s");
			ADD_UIMM(WT_TOMB_OP_SUB_I, "sub.i");
			ADD_MBZ(WT_TOMB_OP_SUB_S, "sub.s");
			ADD_UIMM(WT_TOMB_OP_MUL_I, "mul.i");
			ADD_MBZ(WT_TOMB_OP_MUL_S, "mul.s");
			ADD_UIMM(WT_TOMB_OP_DIV_I, "div.i");
			ADD_MBZ(WT_TOMB_OP_DIV_S, "div.s");
			ADD_UIMM(WT_TOMB_OP_SHL_I, "shl.i");
			ADD_MBZ(WT_TOMB_OP_SHL_S, "shl.s");
			ADD_UIMM(WT_TOMB_OP_SHR_I, "shr.i");
			ADD_MBZ(WT_TOMB_OP_SHR_S, "shr.s");
			ADD_UIMM(WT_TOMB_OP_SAR_I, "sar.i");
			ADD_MBZ(WT_TOMB_OP_SAR_S, "sar.s");
			ADD_UIMM(WT_TOMB_OP_ROR_I, "ror.i");
			ADD_MBZ(WT_TOMB_OP_ROR_S, "ror.s");
			ADD_UIMM(WT_TOMB_OP_SMUL_I, "smul.i");
			ADD_MBZ(WT_TOMB_OP_SMUL_S, "smul.s");
			ADD_UIMM(WT_TOMB_OP_SDIV_I, "sdiv.i");
			ADD_MBZ(WT_TOMB_OP_SDIV_S, "sdiv.s");
			ADD_MBZ(WT_TOMB_OP_MOD, "mod");
			ADD_MBZ(WT_TOMB_OP_SMOD, "smod");
			ADD_UIMM(WT_TOMB_OP_LSOR_I, "lsor.i");
			ADD_ULIMM(WT_TOMB_OP_LSOR_L, "lsor.l");
			ADD_UIMM(WT_TOMB_OP_RSB_I, "rsb.i");
			ADD_ULIMM(WT_TOMB_OP_RSB_L, "rsb.l");
			ADD_MBZ(WT_TOMB_OP_ADDL, "addl");
			ADD_MBZ(WT_TOMB_OP_SUBL, "subl");
			ADD_MBZ(WT_TOMB_OP_MULL, "mull");
			ADD_MBZ(WT_TOMB_OP_DIVL, "divl");
			ADD_MBZ(WT_TOMB_OP_SHLL, "shll");
			ADD_MBZ(WT_TOMB_OP_SHRL, "shrl");
			ADD_MBZ(WT_TOMB_OP_SARL, "sarl");
			ADD_MBZ(WT_TOMB_OP_RORL, "rorl");
			ADD_MBZ(WT_TOMB_OP_XOR, "xor");
			ADD_MBZ(WT_TOMB_OP_BAND, "band");
			ADD_MBZ(WT_TOMB_OP_BOR, "bor");
			ADD_MBZ(WT_TOMB_OP_BNOT, "bnot");
			ADD_MBZ(WT_TOMB_OP_XORL, "xorl");
			ADD_MBZ(WT_TOMB_OP_BANDL, "bandl");
			ADD_MBZ(WT_TOMB_OP_BORL, "borl");
			ADD_MBZ(WT_TOMB_OP_BNOTL, "bnotl");
			ADD_UIMM(WT_TOMB_OP_LSORL_I, "lsorl.i");
			ADD_ULIMM(WT_TOMB_OP_LSORL_L, "lsorl.l");
			ADD_MBZ(WT_TOMB_OP_SMULL, "smull");
			ADD_MBZ(WT_TOMB_OP_SDIVL, "sdivl");
			ADD_MBZ(WT_TOMB_OP_MODL, "modl");
			ADD_MBZ(WT_TOMB_OP_SMODL, "smodl");
			ADD_MBZ(WT_TOMB_OP_NEG, "neg");
			ADD_MBZ(WT_TOMB_OP_NEGL, "negl");
			ADD_MBZ(WT_TOMB_OP_FADD, "fadd");
			ADD_MBZ(WT_TOMB_OP_FSUB, "fsub");
			ADD_MBZ(WT_TOMB_OP_FMUL, "fmul");
			ADD_MBZ(WT_TOMB_OP_FDIV, "fdiv");
			ADD_MBZ(WT_TOMB_OP_FADDL, "faddl");
			ADD_MBZ(WT_TOMB_OP_FSUBL, "fsubl");
			ADD_MBZ(WT_TOMB_OP_FMULL, "fmull");
			ADD_MBZ(WT_TOMB_OP_FDIVL, "fdivl");
			ADD_MBZ(WT_TOMB_OP_FNEG, "fneg");
			ADD_MBZ(WT_TOMB_OP_FNEGL, "fnegl");
			ADD_MBZ(WT_TOMB_OP_FMOD, "fmod");
			ADD_MBZ(WT_TOMB_OP_FMODL, "fmodl");
			ADD_MBZ(WT_TOMB_OP_FABS, "fabs");
			ADD_MBZ(WT_TOMB_OP_FABSL, "fabsl");

			ADD_MBZ(WT_TOMB_OP_CMP, "cmp");
			ADD_MBZ(WT_TOMB_OP_SCMP, "scmp");
			ADD_MBZ(WT_TOMB_OP_CMPL, "cmpl");
			ADD_MBZ(WT_TOMB_OP_SCMPL, "scmpl");
			ADD_MBZ(WT_TOMB_OP_FCMP, "fcmp");
			ADD_MBZ(WT_TOMB_OP_FCMPL, "fcmpl");

			ADD_MBZ(WT_TOMB_OP_QLT, "qlt");
			ADD_MBZ(WT_TOMB_OP_QLE, "qle");
			ADD_MBZ(WT_TOMB_OP_QGT, "qgt");
			ADD_MBZ(WT_TOMB_OP_QGE, "qge");
			ADD_MBZ(WT_TOMB_OP_QEQ, "qeq");
			ADD_MBZ(WT_TOMB_OP_QNE, "qne");

			ADD_MBZ(WT_TOMB_OP_LNOT, "lnot");
			ADD_MBZ(WT_TOMB_OP_LAND, "land");
			ADD_MBZ(WT_TOMB_OP_LOR, "lor");
			ADD_MBZ(WT_TOMB_OP_LXOR, "lxor");
			ADD_MBZ(WT_TOMB_OP_LDB, "ldb");
			ADD_MBZ(WT_TOMB_OP_STB, "stb");
			ADD_MBZ(WT_TOMB_OP_LDH, "ldh");
			ADD_MBZ(WT_TOMB_OP_STH, "sth");
		}

		WTR_MICROPROGRAMS(__WTR_FILL_MICROPROGS);
		disasm->num_microprogs = WTR_NUM_MICROPROGRAMS;
	}

	return disasm;
}

void Disassembler_Release(struct Disassembler* const disasm)
{
	free(disasm);
}

const wt_u16* Disassembler_Disassemble(struct Disassembler* disasm, wt_bool vermatch,
	wt_u32 addr, const wt_u16* code, wt_u32 size, struct labelmap* labels, wt_bo bo)
{
	const wt_u16* const end = code + (size/sizeof(*code));

	disasm->a_byteorder     = bo;
	disasm->a_version_match = vermatch;

	while(code<end)
	{
		struct InstructionDecoder* const id  = disasm->decoder + WT_TOMB_GETOPCODE(disasm_read(disasm, *code));
		const char* const                lab = labels ? labelmap_find(labels, addr) : NULL;

		if(NULL!=lab)
		{
			printf(INDENT "%08X <%s>:\n", addr, lab);
		}

		if(id->mnemonic && id->decode(id, disasm, addr, code, labels))
		{
			code += id->width;
			addr += id->width * sizeof(*code);
		}
		else
		{
			return code;
		}
	}

	return NULL;
}
