//================================================================================================
// $Id: cgsection.cpp 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator section
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "cgsection.h"
#include "cgfile.h"
#include "codegen.h"
#include "parserext.h"
#include "wtruntime/hash.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_recordtype.h"
#include "wtparser/ast/ast_recordfield.h"
#include "wtparser/ast/ast_operatordeclaration.h"
#include "wtparser/ast/ast_datatype.h"
#include "wtparser/ast/ast_datatypecons.h"
#include "wtparser/ast/ast_functiontype.h"
#include "wtparser/ast/ast_enumtype.h"
#include "wtparser/ast/ast_enumerator.h"

namespace wtc
{
namespace codegen
{

const std::string Section::EmptyString("");

wt_u32 MetadataSection::GetSize() const
{
	wt_u32 sum = sizeof(wtb_metahdr) + sizeof(wtb_meta)*mMetadata.size();

	for(metadata_map_t::const_iterator it=mMetadata.begin(),end=mMetadata.end();it!=end;++it)
	{
		sum += it->first.length() + 1;
		sum += it->second.length() + 1;
	}

	return sum;
}

void MetadataSection::Write(File & file) const
{
	{
		wtb_metahdr mhdr;
		mhdr.metacount = mMetadata.size();
		mhdr.reserved  = 0;
		file.Write(mhdr);
	}

	wt_u32 offs = sizeof(wtb_metahdr) + sizeof(wtb_meta)*mMetadata.size();

	for(metadata_map_t::const_iterator it=mMetadata.begin(),end=mMetadata.end();it!=end;++it)
	{
		wtb_meta meta;
		meta.meta_key_offs = offs;
		offs += it->first.length() + 1;
		meta.meta_val_offs = offs;
		offs += it->second.length() + 1;
		file.Write(meta);
	}

	for(metadata_map_t::const_iterator it=mMetadata.begin(),end=mMetadata.end();it!=end;++it)
	{
		file.Write(it->first);
		file.Write(it->second);
	}
}

namespace integrity
{
class HashGenerator : public PostDataGenerator
{
	wt_u16 mHashAlg;
	wt_u32 mHashLen;
	wt_bo  mByteOrder;

public:
	HashGenerator(wtc::ProblemHandler & probhan, wt_u16 hash_alg,
		wt_u32 hash_len, wt_bo bo)
		: PostDataGenerator(probhan)
		, mHashAlg(hash_alg)
		, mHashLen(hash_len)
		, mByteOrder(bo)
	{
		// ok
	}

	void Generate(std::vector<wt_u8> & out, const std::vector<wt_u8> & data)
	{
		out.resize(mHashLen);

		if(WTE_OK!=wtr_hash_calc(&out[0], mHashAlg, &data[0], data.size(),
			mByteOrder))
		{
			mProbHan.AddError("<internal>", 0, "Hash cannot be calculated");
		}
	}
};
} // integrity

void IntegritySection::Write(File & file) const
{
	wtb_integhdr ihdr;
	ihdr.algorithm = mHashAlg;
	ihdr.reserved = 0;
	file.Write(ihdr);
	file.AddPostData(new integrity::HashGenerator(mProbHan, mHashAlg,
		mHashSize, mByteOrder), mHashSize);
}

bool MetadataSection::Add(const std::string & key, const std::string & value)
{
	bool result = mMetadata.end()!=mMetadata.find(key);
	mMetadata[key] = value;
	return result;
}

wt_u32 BosSection::MakeSize(wt_u32 n)
{
	if(0==n) { return 0; }
	if(0x00000100>n) { return 1+1; }
	if(0x00010000>n) { return 1+2; }
	return 1+4;
}

wt_u32 BosSection::GetSize() const
{
	wt_u32 sum = 0;

	sum += MakeSize(mCount2);
	sum += MakeSize(mCount4);
	sum += MakeSize(mCount8);
	sum += MakeSize(mCountS);

	return sum;
}

void BosSection::WriteSize(File & file, wt_u32 n, wt_u8 base_bos)
{
	if(0==n)
	{
		return;
	}
	else
	if(0x00000100>n)
	{
		file.Write(static_cast<wt_u8>(base_bos+0));
		file.Write(static_cast<wt_u8>(n));
	}
	else
	if(0x00010000>n)
	{
		file.Write(static_cast<wt_u8>(base_bos+1));
		file.Write(static_cast<wt_u16>(n));
	}
	else
	{
		file.Write(static_cast<wt_u8>(base_bos+2));
		file.Write(n);
	}
}

void BosSection::Write(File & file) const
{
	WriteSize(file, mCount8, WTBBOS_SWAP64_8);
	WriteSize(file, mCount4, WTBBOS_SWAP32_8);
	WriteSize(file, mCount2, WTBBOS_SWAP16_8);
	WriteSize(file, mCountS, WTBBOS_NOSWAP_8);
}

wt_u32 DataSection::GetSize() const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to DataSection::GetSize()");
		return 0;
	}

	wt_u32 size = 0;

	size += mData2.size() * sizeof(wt_u16);
	size += mData4.size() * sizeof(wt_u32);
	size += mData8.size() * sizeof(wt_u64);

	for(data_mapping_t::const_iterator it=mDataS.begin(),end=mDataS.end();it!=end;++it)
	{
		size += it->first.length() + 1;
	}

	return size;
}

void DataSection::Write(File & file) const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to DataSection::Write()");
		return;
	}

	for(data_mapping_t::const_iterator it=mData8.begin(),end=mData8.end();it!=end;++it)
	{
		file.Write(*reinterpret_cast<const wt_u64*>(it->first.c_str()));
	}

	for(data_mapping_t::const_iterator it=mData4.begin(),end=mData4.end();it!=end;++it)
	{
		file.Write(*reinterpret_cast<const wt_u32*>(it->first.c_str()));
	}

	for(data_mapping_t::const_iterator it=mData2.begin(),end=mData2.end();it!=end;++it)
	{
		file.Write(*reinterpret_cast<const wt_u16*>(it->first.c_str()));
	}

	for(data_mapping_t::const_iterator it=mDataS.begin(),end=mDataS.end();it!=end;++it)
	{
		file.Write(it->first);
	}
}

BosSection* DataSection::Compile()
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to DataSection::Compile()");
		return 0;
	}

	wt_u32 string_size = 0;
	wt_u32 addr = GetAddr();

	for(data_mapping_t::iterator it=mData8.begin(),end=mData8.end();it!=end;addr+=sizeof(wt_u64),++it)
	{
		it->second = addr;
	}

	for(data_mapping_t::iterator it=mData4.begin(),end=mData4.end();it!=end;addr+=sizeof(wt_u32),++it)
	{
		it->second = addr;
	}

	for(data_mapping_t::iterator it=mData2.begin(),end=mData2.end();it!=end;addr+=sizeof(wt_u16),++it)
	{
		it->second = addr;
	}

	for(data_mapping_t::iterator it=mDataS.begin(),end=mDataS.end();it!=end;++it)
	{
		it->second = addr;
		wt_u32 act_size = it->first.length() + 1;
		addr += act_size;
		string_size += act_size;
	}

	BosSection* const bos = new BosSection(mProbHan, mBosName, GetAddr(),
		GetName(), mData2.size(), mData4.size(), mData8.size(), string_size);

	if(!bos)
	{
		mProbHan.AddError("<internal>", 0,
			"Could not allocate memory for BOS section");
		return 0;
	}

	mCompiled = true;
	return bos;
}

void DataSection::Add(data_mapping_t & mapping, const std::string & data)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to DataSection::Add()");
		return;
	}

	mapping[data] = 0;
}

wt_u32 DataSection::Find(const data_mapping_t & mapping, const std::string & data) const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to DataSection::Find()");
		return 0;
	}

	data_mapping_t::const_iterator it = mapping.find(data);

	if(it!=mapping.end())
	{
		return it->second;
	}

	mProbHan.AddError("<internal>", 0, "Data mapping error");
	return 0;
}

void ExternalsSection::_Add(map_name_offs_t & mapping, const std::string & name)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[name] = 0;
}

void ExternalsSection::_Add(map_name_offssig_t & mapping, wtast::FunctionSignature* sig)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[sig->GetCanonicalName(false, true)] = std::make_pair(0, sig);
}

void ExternalsSection::_Add(map_name_offstype_t & mapping, wtast::RecordType* rec)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[rec->GetName()] = std::make_pair(rec->GetTypeId() & WTTM_USERTYPEID, rec);
}

void ExternalsSection::_Add(map_name_offstype_t & mapping, wtast::DataType* dt)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[dt->GetName()] = std::make_pair(dt->GetTypeId() & WTTM_USERTYPEID, dt);
}

void ExternalsSection::_Add(map_name_offstype_t & mapping, wtast::EnumType* et)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[et->GetName()] = std::make_pair(et->GetTypeId() & WTTM_USERTYPEID, et);
}

void ExternalsSection::_Add(map_name_offstype_t & mapping, wtast::FunctionType* ft)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[ft->GetName()] = std::make_pair(ft->GetTypeId() & WTTM_USERTYPEID, ft);
}

void ExternalsSection::_Add(map_name_op_t & mapping, wtast::OperatorDeclaration* op)
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::_Add()");
		return;
	}

	mapping[op->GetOpName()] = op;
}

const wt_u32 ExternalsSection::BadAddr = ~(uint32_t)0;
const wt_u32 ExternalsSection::BadOffs = ~(uint32_t)0;

wt_u32 ExternalsSection::_GetNameAddr(const map_name_offs_t & mapping,
	const std::string & name) const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::GetNameAddr()");
		return BadAddr;
	}

	map_name_offs_t::const_iterator it = mapping.find(name);

	if(mapping.end()!=it)
	{
		return it->second;
	}

	mProbHan.AddError("?", 0, std::string("Could not resolve '") + name +
		"'");
	return BadAddr;
}

wt_u32 ExternalsSection::_GetNameAddr(const map_name_offssig_t & mapping,
	const std::string & name) const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::GetNameAddr()");
		return BadAddr;
	}

	map_name_offssig_t::const_iterator it = mapping.find(name);

	if(mapping.end()!=it)
	{
		return it->second.first;
	}

	mProbHan.AddError("?", 0, std::string("Could not resolve '") + name +
		"'");
	return BadAddr;
}

wt_u32 ExternalsSection::_GetNameOffs(const map_name_offstype_t & mapping,
	const std::string & name) const
{
	if(!mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Invalid access to ExternalsSection::GetNameOffs()");
		return BadOffs;
	}

	map_name_offstype_t::const_iterator it = mapping.find(name);

	if(mapping.end()!=it)
	{
		return it->second.first;
	}

	mProbHan.AddError("?", 0, std::string("Could not resolve '") + name +
		"'");
	return BadOffs;
}

void ExternalsSection::Compile()
{
	if(mCompiled)
	{
		mProbHan.AddError("<internal>", 0,
			"Duplicate call to ExternalsSection::Compile()");
		return;
	}

	mCompiled = true;

	wt_u32 base = GetAddr();
	wt_u32 addr = base;
	wt_u32 numt = 0;

	std::set<wt_u32> indices;

	for(map_name_offstype_t::iterator it=mNameToType.begin(),end=mNameToType.end();it!=end;++it)
	{
		wt_u32 type_id    = it->second.second->GetTypeId();
		wt_u32 type_index = type_id & WTTM_USERTYPEID;

		if(!(type_id & WTTF_LOCAL))
		{
			mProbHan.AddError("?", 0, "External is not a local user type.");
			continue;
		}

		if(indices.count(type_index))
		{
			mProbHan.AddError("?", 0, "Reused external type index.");
			continue;
		}

		if(numt<=type_index)
		{
			numt = type_index+1;
		}

		it->second.first = type_index;
	}

	indices.clear();
	addr = base + numt*sizeof(wt_handle);

	for(map_name_offssig_t::iterator it=mNameToPFN.begin(),end=mNameToPFN.end();it!=end;++it)
	{
		it->second.first = addr;
		addr += sizeof(wt_pfn);
	}

	for(map_name_offs_t::iterator it=mNameToExc.begin(),end=mNameToExc.end();it!=end;++it)
	{
		it->second = addr;
		addr += sizeof(wt_exception);
	}

	mCompiledSize = addr - base;
}

void CodeSection::Write(File & file) const
{
	for(instructions_t::const_iterator it=mInstructions.begin(),end=mInstructions.end();it!=end;++it)
	{
		file.Write(*it);
	}
}

wt_u32 CodeSection::GetFunctionPC(const std::string & name) const
{
	map_func_pc_t::const_iterator it = mFunctions.find(name);

	if(mFunctions.end()!=it)
	{
		return it->second;
	}

	return 0; // invalid PC
}

void GenericImportExportSection::Write(File & file) const
{
	wtb_symhdr hdr;
	hdr.symcount = mSymbols.size();
	hdr.reserved = 0;

	file.Write(hdr);

	wt_u32 base_offs = sizeof(wtb_symhdr) + sizeof(wtb_sym)*mSymbols.size();

	for(symvec_t::const_iterator it=mSymbols.begin(),end=mSymbols.end();it!=end;++it)
	{
		wtb_sym sym = *it;
		sym.sym_name += base_offs;

		if(sym.sym_offs)
		{
			sym.sym_offs += base_offs;
		}

		file.Write(sym);
	}

	for(datavec_t::const_iterator it=mData.begin(),end=mData.end();it!=end;++it)
	{
		switch(it->type)
		{
		case Data::T_U8:     file.Write(it->v_u8);     break;
		case Data::T_U32:    file.Write(it->v_u32);    break;
		case Data::T_STRING: file.Write(it->v_string); break;
		default:
			mProbHan.AddError("<internal>", 0,
				"Unknown data type in ImportSection::Write()");
		}
	}
}

void GenericImportExportSection::PlaceString(const std::string & str)
{
	PlaceString(str, Padding);
}

void GenericImportExportSection::PlaceString(const std::string & str, wt_u32 padding)
{
	std::size_t len = str.length() + 1;
	std::size_t pad = (len + padding - 1) & ((~padding) + 1);

	mData.push_back(Data(str));
	for(;len<pad;++len)
	{
		mData.push_back(Data(wt_u8(0)));
	}

	mRelSize += pad;
}

void GenericImportExportSection::PlaceFuncSig(wtast::FunctionSignature* sig)
{
	wt_u32 argc = wt_u32(sig->GetArgc());
	wt_u32 retc = wt_u32(sig->GetRetc());

	mData.push_back(Data(argc));

	for(wt_u32 i=0;i<argc;++i)
	{
		mData.push_back(sig->GetArg(i)->GetType()->GetTypeId());
	}

	mData.push_back(Data(retc));

	for(wt_u32 i=0;i<retc;++i)
	{
		mData.push_back(sig->GetRet(i)->GetTypeId());
	}

	if(sig->IsPure())
	{
		mData.push_back(Data(wt_u8(1)));
	}
	else
	{
		mData.push_back(Data(wt_u8(0)));
	}

	// padding

	mData.push_back(Data(wt_u8(0)));
	mData.push_back(Data(wt_u8(0)));
	mData.push_back(Data(wt_u8(0)));

	mRelSize += (argc + retc + 2)*sizeof(wt_u32) + 4;
}

void GenericImportExportSection::PlaceRecord(wtast::RecordType* rec)
{
	wt_u32 fldc = wt_u32(rec->GetNumberOfFields());

	mData.push_back(Data(fldc));

	for(wt_u32 i=0;i<fldc;++i)
	{
		mData.push_back(rec->GetField(i)->GetType()->GetTypeId());
	}

	mRelSize += (fldc + 1)*sizeof(wt_u32);

	for(wt_u32 i=0;i<fldc;++i)
	{
		wt_u32 padding = 1;

		if(i+1==fldc)
		{
			padding = Padding;
		}

		PlaceString(rec->GetField(i)->GetName(), padding);
	}
}

void GenericImportExportSection::PlaceDataCons(wtast::DataTypeCons* cons)
{
	PlaceString(cons->GetName());

	wt_u32 typn = wt_u32(cons->GetNumberOfTypes());

	mData.push_back(Data(typn));

	for(wt_u32 i=0;i<typn;++i)
	{
		mData.push_back(cons->GetType(i)->GetTypeId());
	}

	mRelSize += (typn + 1)*sizeof(wt_u32);
}

void GenericImportExportSection::SYM_FunctionType(wtast::FunctionType* ft, wt_u32 sym_info)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_FUNCTIONTYPE;
	sym.sym_name = mRelSize;

	PlaceString(ft->GetName());

	sym.sym_offs = mRelSize;

	PlaceFuncSig(ft->GetFunctionSignature());

	sym.sym_info = sym_info;
	mSymbols.push_back(sym);
}

void GenericImportExportSection::SYM_DataType(wtast::DataType* dt, wt_u32 sym_info)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_DATATYPE;
	sym.sym_name = mRelSize;

	PlaceString(dt->GetName());

	wt_u32 defcons = 0;
	if(dt->GetDefaultCons())
	{
		defcons = dt->GetDefaultCons()->GetConsId();
	}

	sym.sym_offs = mRelSize;
	sym.sym_info = sym_info;
	mData.push_back(wt_u32(dt->GetNumberOfConses()));
	mData.push_back(Data(defcons));
	mRelSize += sizeof(wt_u32)*2;
	for(std::size_t i=0,n=dt->GetNumberOfConses();i!=n;++i)
	{
		PlaceDataCons(dt->GetCons(i));
	}

	mSymbols.push_back(sym);
}

void GenericImportExportSection::SYM_EnumType(wtast::EnumType* et, wt_u32 sym_info)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_ENUM;
	sym.sym_name = mRelSize;

	PlaceString(et->GetName());

	sym.sym_offs = mRelSize;
	sym.sym_info = sym_info;
	const wt_u32 ec = wt_u32(et->GetNumberOfEnumerators());
	mData.push_back(ec);
	mRelSize += sizeof(wt_u32);

	for(wt_u32 i=0,n=ec;i!=n;++i)
	{
		wt_u32 padding = 1;

		if(i+1==ec)
		{
			padding = Padding;
		}

		PlaceString(et->GetEnumerator(i)->GetName(), padding);
	}

	mSymbols.push_back(sym);
}

void GenericImportExportSection::SYM_Operator(wtast::OperatorDeclaration* op)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_OPERATOR;
	sym.sym_name = mRelSize;

	PlaceString(op->GetOpName());

	wt_u32 assoc = WTBSYM_OP_RIGHTASSOC;
	if(op->IsLeftAssociative())
	{
		assoc = WTBSYM_OP_LEFTASSOC;
	}

	sym.sym_offs = 0;
	sym.sym_info = WTBSYM_CR_OP(assoc, op->GetPrecedence());
	mSymbols.push_back(sym);
}

void ImportSection::AddFunctionImport(wtast::FunctionSignature* sig, wt_u32 offs)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_FUNCTION;
	sym.sym_name = mRelSize;

	PlaceString(sig->GetName());

	sym.sym_offs = mRelSize;

	PlaceFuncSig(sig);

	sym.sym_info = offs;
	mSymbols.push_back(sym);
}

void ImportSection::AddExceptionImport(const std::string & exc_name, wt_u32 offs)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_EXCEPTION;
	sym.sym_name = mRelSize;
	sym.sym_offs = 0;
	sym.sym_info = offs;

	PlaceString(exc_name);

	mSymbols.push_back(sym);
}

void ImportSection::AddTypeImport(wtast::Type* type, wt_u32 info)
{
	wtb_sym sym;

	switch(type->GetKind())
	{
	case wtast::AstNode::RECORDTYPE:
		{
			wtast::RecordType* rec        = static_cast<wtast::RecordType*>(type);
			const bool         is_private = rec->IsPrivate();
			const bool         is_limited = rec->IsLimited();

			if(is_limited)
			{
				sym.sym_type = WTBSYMT_LIMITEDRECORD;
			}
			else
			if(is_private)
			{
				sym.sym_type = WTBSYMT_PRIVATERECORD;
			}
			else
			{
				sym.sym_type = WTBSYMT_RECORD;
			}

			sym.sym_name = mRelSize;
			PlaceString(rec->GetName());

			if(is_private || is_limited)
			{
				sym.sym_offs = 0;
			}
			else
			{
				sym.sym_offs = mRelSize;
				PlaceRecord(rec);
			}

			sym.sym_info = info;
			mSymbols.push_back(sym);
		}
		break;

	case wtast::AstNode::DATATYPE:
		{
			SYM_DataType(static_cast<wtast::DataType*>(type), info);
		}
		break;

	case wtast::AstNode::ENUMTYPE:
		{
			SYM_EnumType(static_cast<wtast::EnumType*>(type), info);
		}
		break;

	case wtast::AstNode::FUNCTIONTYPE:
		{
			SYM_FunctionType(static_cast<wtast::FunctionType*>(type), info);
		}
		break;

	default:
		mProbHan.AddError("?", 0, "Unhandled type import.");
	}
}

void ImportSection::AddOperatorImport(wtast::OperatorDeclaration* op)
{
	SYM_Operator(op);
}

void ExportSection::AddExceptionExport(const std::string & exc_name)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_EXCEPTION;
	sym.sym_name = mRelSize;
	sym.sym_offs = 0;
	sym.sym_info = 0;

	PlaceString(exc_name);

	mSymbols.push_back(sym);
}

void ExportSection::AddFunctionExport(wtast::FunctionSignature* sig, wt_u32 pc)
{
	wtb_sym sym;

	sym.sym_type = WTBSYMT_FUNCTION;
	sym.sym_name = mRelSize;

	PlaceString(sig->GetName());

	sym.sym_offs = mRelSize;

	PlaceFuncSig(sig);

	sym.sym_info = pc;
	mSymbols.push_back(sym);
}

void ExportSection::AddTypeExport(wtast::Type* type, wt_u32 sym_info)
{
	wtb_sym sym;

	switch(type->GetKind())
	{
	case wtast::AstNode::RECORDTYPE:
		{
			wtast::RecordType* rec        = static_cast<wtast::RecordType*>(type);
			const bool         is_private = rec->IsPrivate();
			const bool         is_limited = rec->IsLimited();

			if(is_limited)
			{
				sym.sym_type = WTBSYMT_LIMITEDRECORD;
			}
			else
			if(is_private)
			{
				sym.sym_type = WTBSYMT_PRIVATERECORD;
			}
			else
			{
				sym.sym_type = WTBSYMT_RECORD;
			}

			sym.sym_name = mRelSize;
			PlaceString(rec->GetName());

			if(is_private || is_limited)
			{
				sym.sym_offs = 0;
			}
			else
			{
				sym.sym_offs = mRelSize;
				PlaceRecord(rec);
			}

			sym.sym_info = sym_info;
			mSymbols.push_back(sym);
		}
		break;

	case wtast::AstNode::DATATYPE:
		{
			SYM_DataType(static_cast<wtast::DataType*>(type), sym_info);
		}
		break;

	case wtast::AstNode::ENUMTYPE:
		{
			SYM_EnumType(static_cast<wtast::EnumType*>(type), sym_info);
		}
		break;

	case wtast::AstNode::FUNCTIONTYPE:
		{
			SYM_FunctionType(static_cast<wtast::FunctionType*>(type), sym_info);
		}
		break;

	default:
		mProbHan.AddError("?", 0, "Unhandled type export.");
	}
}

void ExportSection::AddOperatorExport(wtast::OperatorDeclaration* op)
{
	SYM_Operator(op);
}

wt_u32 DLabelSection::GetSize() const
{
	wt_u32 sum = sizeof(wtb_dlabelhdr) + sizeof(wtb_dlabel)*mMapAddrToLabel.size();

	for(map_addr_label_t::const_iterator it=mMapAddrToLabel.begin(),end=mMapAddrToLabel.end();it!=end;++it)
	{
		sum += it->second.length() + 1;
	}

	return sum;
}

void DLabelSection::Write(File & file) const
{
	wt_u32 offs = sizeof(wtb_dlabelhdr) + sizeof(wtb_dlabel)*mMapAddrToLabel.size();

	wtb_dlabelhdr hdr;
	hdr.labelcount = mMapAddrToLabel.size();
	hdr.reserved   = 0;

	file.Write(hdr);

	for(map_addr_label_t::const_iterator it=mMapAddrToLabel.begin(),end=mMapAddrToLabel.end();it!=end;++it)
	{
		wtb_dlabel dlab;
		dlab.addr = it->first;
		dlab.lab_offs = offs;
		file.Write(dlab);

		offs += it->second.length() + 1;
	}

	for(map_addr_label_t::const_iterator it=mMapAddrToLabel.begin(),end=mMapAddrToLabel.end();it!=end;++it)
	{
		file.Write(it->second);
	}
}

} // codegen
} // wtc
