#include "config_private.h"
#include "ctx.h"
#include "profile.h"
#include "stdlib.h"
#include "objects.h"
#include "function.h"
#include "funcsig.h"
#include "processor.h"
#include <string.h>

#define SBUF_MINSIZE  32
#define SBUF_SIZE     1024

#define STACK_MINSIZE 128
#define STACK_MAXSIZE 0x01000000

#define TOS_MINSIZE   32
#define TOS_MAXSIZE   65536
#define TOS_LIMIT     12 /* 24 ? */

#define RTOS_MINSIZE  32
#define RTOS_MAXSIZE  8192
#define RTOS_LIMIT    24

static void wtr_ctx_setup(struct wtr_context* const ctx, wt_u32 pc,
	struct wtr_program* const prog, struct wtr_function* const func,
	WTR_PFN_RETURNACTION return_action);

static wt_error wtr_ctx_invoke_prog(struct wtr_context* const ctx,
	struct wtr_program* const prog, wt_u32 pc,
	WTR_PFN_RETURNACTION return_action);

static void decrement_called_refcounts(struct wtr_context* const ctx);

wt_error wtrContextCreate(struct wtr_context** const p_context,
	wt_profile profile, wt_u32 stack_size, wt_u32 sbuf_size, wt_u32 flags)
{
	wt_error            err;
	struct wtr_context* ctx;

	size_t size = sizeof(*ctx);

	if(NULL==p_context)
	{
		return WTE_BADARG;
	}

	if(0==sbuf_size)
	{
		sbuf_size = SBUF_SIZE;
	}
	else
	if(SBUF_MINSIZE>sbuf_size)
	{
		sbuf_size = SBUF_MINSIZE;
	}

	if(STACK_MINSIZE>stack_size)
	{
		stack_size = STACK_MINSIZE;
	}
	else
	if(STACK_MAXSIZE<stack_size)
	{
		stack_size = STACK_MAXSIZE;
	}

	size += sbuf_size;

	if(NULL==(ctx=wtr_profile_sys_malloc(profile, size)))
	{
		return WTE_NOMEM;
	}

	memset(ctx, 0, sizeof(*ctx));

	ctx->profile   = profile;
	ctx->run       = WTRR_INIT;
	ctx->sbuf      = (char*)(ctx + 1);
	ctx->sbuf_size = sbuf_size;

	if(!wtr_dynstack_init(&ctx->tos, profile, TOS_MINSIZE, TOS_MAXSIZE,
		TOS_LIMIT))
	{
		err = WTE_NOMEM;
		goto error_1;
	}

	if(!wtr_dynstack_init(&ctx->rtos, profile, RTOS_MINSIZE, RTOS_MAXSIZE,
		RTOS_LIMIT))
	{
		err = WTE_NOMEM;
		goto error_2;
	}

	if(NULL==(ctx->stack_alloc=wtr_profile_vm_malloc(profile, stack_size)))
	{
		err = WTE_NOMEM;
		goto error_3;
	}

	/* split allocated stack to CS and DS */
	{
		wt_u32 dss;

		wt_u32 css =
			(stack_size * profile->ratio_css)
			/
			(profile->ratio_css + profile->ratio_dss);

		css = (css + 7) & (~7);

		dss = stack_size - css;

		while(24>dss)
		{
			dss += 8;
			css += 8;
		}

		ctx->cs_lo = (wt_u8*)ctx->stack_alloc;
		ctx->cs_hi = ctx->cs_lo + css;
		ctx->ds_lo = ctx->cs_hi;
		ctx->ds_hi = ctx->ds_lo + dss;

		/* preserve space for an exception to be able to handle stack */
		/* overflow error */
		ctx->ds_lo += sizeof(wt_exception);
	}

	ctx->csp = ctx->cs_hi;
	ctx->dsp = ctx->ds_hi;

	*p_context = ctx;
	return WTE_OK;

error_3:;
	wtr_dynstack_free(&ctx->rtos);
error_2:;
	wtr_dynstack_free(&ctx->tos);
error_1:;
	wtr_profile_sys_free(profile, ctx);
	return err;
}

wt_error wtrContextDestroy(struct wtr_context* const context)
{
	if(WTRR_INIT!=context->run)
	{
		return WTE_BADCTX;
	}

	if(0!=context->tos.sp || 0!=context->rtos.sp)
	{
		return WTE_UCLEAN;
	}

	wtr_profile_vm_free(context->profile, context->stack_alloc);

	if(NULL!=context->value_array)
	{
		wtr_profile_sys_free(context->profile, context->value_array);
	}

	wtr_dynstack_free(&context->rtos);
	wtr_dynstack_free(&context->tos);

	wtr_profile_sys_free(context->profile, context);

	return WTE_OK;
}

static void ctxcleanup_return_action(struct wtr_context* const ctx)
{
	decrement_called_refcounts(ctx);
}

wt_error wtrContextCleanup(struct wtr_context* const context)
{
	if(WTRR_INIT!=context->run)
	{
		return WTE_BADCTX;
	}

	wtr_ctx_setup(context, context->profile->pc_tcl_leave, NULL, NULL,
		ctxcleanup_return_action);

	context->retc = 0;
	context->run = WTRR_RUN;
	return WTE_OK;
}

wt_error wtrContextSetUserPointer(struct wtr_context* const context,
	void* ptr)
{
	context->user_pointer = ptr;
	return WTE_OK;
}

wt_error wtrContextGetUserPointer(struct wtr_context* const context,
	void** pptr)
{
	if(NULL!=pptr)
	{
		*pptr = context->user_pointer;
	}

	return WTE_OK;
}

static void decrement_called_refcounts(struct wtr_context* const ctx)
{
	if(NULL!=ctx->called_func)
	{
		--ctx->called_func->refcount;
		ctx->called_func = NULL;
	}

	if(NULL!=ctx->called_prog)
	{
		--ctx->called_prog->refcount;
		ctx->called_prog = NULL;
	}
}

static void invoke_return_action(struct wtr_context* const ctx)
{
	decrement_called_refcounts(ctx);
}

wt_error wtrContextInvokeProgram(struct wtr_context* const context,
	wt_program program)
{
	return wtr_ctx_invoke_prog(context, program,
		program->pc_base+program->pc_main, invoke_return_action);
}

wt_error wtrContextInvokeFunction(struct wtr_context* const context,
	wt_function func, wt_u32 argc, wt_value* argv)
{
	struct wtr_funcsig* const sig = func->funcsig;

	wt_error err;
	wt_u32   i;
	wt_u32   dneed = sizeof(func->pfn);

	if(WTRR_INIT!=context->run)
	{
		return WTE_BADCTX;
	}

	if(sig->argc!=argc)
	{
		return WTE_BADARG;
	}

	for(i=0;i<argc;++i)
	{
		const wt_u32 type_id = argv[i].type;

		if(type_id!=sig->argv[i])
		{
			return WTE_BADARG;
		}

		dneed += wtr_sizeof(type_id);
	}

	if(WTR_CTX_DS_SPACE(context)<dneed)
	{
		return WTE_NOSTK;
	}

	if(0!=wtr_ctx_push_values(context, argc, argv))
	{
		context->dsp = context->ds_hi;
		return WTE_NOSTK;
	}

	WTR_CTX_DS_PUSH(context, wt_pfn, func->pfn);

	wtr_ctx_setup(context, context->profile->pc_cfp_leave, func->program,
		func, invoke_return_action);

	context->retc = sig->retc;
	context->run = WTRR_RUN;
	return WTE_OK;
}

wt_error wtrContextPopReturnValues(struct wtr_context* const context,
	wt_u32* const retc, wt_u32 retmax, wt_value* retv)
{
	wt_exception exc;

	if(WTRR_RETURNED!=context->run)
	{
		return WTE_BADCTX;
	}

	if(NULL==retc || retmax<context->retc || (0<retmax && NULL==retv))
	{
		return WTE_BADARG;
	}

	*retc = context->retc;

	if(0!=(exc=wtr_ctx_pop_values(context, context->retc, retv,
		((0<context->retc) ? context->called_func->funcsig->retv : NULL))))
	{
		// TODO
	}

	context->run = WTRR_INIT;

	return WTE_OK;
}

enum wtr_runresult wtrContextRun(struct wtr_context* const context,
	wt_u32 quantum)
{
	if(WTRR_QUANTUM==context->run)
	{
		context->run = WTRR_RUN;
	}
	else
	{
		switch(context->run)
		{
		case WTRR_RUN:
		case WTRR_RAISE:
			break;

		default:
			return context->run;
		}
	}

	return wtr_processor_tomb_run(context->profile, context, quantum);
}

wt_error wtrContextRaiseException(struct wtr_context* const context,
	wt_exception exception_id)
{
	if(WTRR_INIT==context->run)
	{
		return WTE_BADCTX;
	}

	context->last_exception = exception_id;
	context->run = WTRR_RAISE;

	return WTE_OK;
}

wt_error wtrContextGetException(struct wtr_context* const context,
	wt_exception* const out_exception_id)
{
	if(WTRR_EXCEPTION!=context->run)
	{
		return WTE_BADCTX;
	}

	if(NULL==out_exception_id)
	{
		return WTE_BADARG;
	}

	*out_exception_id = context->last_exception;
	context->run = WTRR_INIT;

	return WTE_OK;
}

wt_error wtrContextGetExtCallParams(struct wtr_context* const context,
	struct wtr_extcallinfo** const p_callinfo, wt_value** const p_argv)
{
	if(WTRR_EXTCALL!=context->run)
	{
		return WTE_BADCTX;
	}

	if(NULL==p_callinfo || NULL==p_argv)
	{
		return WTE_BADARG;
	}

	*p_callinfo = &context->extcallinfo;
	*p_argv     = context->value_array;

	return WTE_OK;
}

wt_error wtrContextGetIOCallParams(struct wtr_context* const context,
	enum wtr_iocall* const iocall, wt_u32* const argc, wt_u32 argmax,
	wt_value* argv)
{
	wt_error err;

	if(WTRR_IOCALL!=context->run)
	{
		return WTE_BADCTX;
	}

	if(NULL==iocall || NULL==argc || (0<argmax && NULL==argv))
	{
		return WTE_BADARG;
	}

	if(argmax<context->iocall_argc)
	{
		return WTE_NOFIT;
	}

	*iocall = context->last_iocall;
	*argc   = context->iocall_argc;

	memcpy(argv, context->iocall_argv,
		sizeof(context->iocall_argv[0])*context->iocall_argc);

	return WTE_OK;
}

wt_error wtrContextPushReturnValues(struct wtr_context* const context,
	wt_u32 retc, wt_value* retv)
{
	wt_u32       i;
	wt_exception exc;

	if(WTRR_IOCALL==context->run)
	{
		WTR_PFN_IOCALLNEXT iocall_next;

		if(retc!=context->iocall_retc || (0<retc && NULL==retv))
		{
			return WTE_BADARG;
		}

		for(i=0;i<retc;++i)
		{
			if(retv[i].type!=context->iocall_retv[i])
			{
				return WTE_BADARG;
			}
		}

		if(0!=(exc=wtr_ctx_push_values(context, retc, retv)))
		{
			context->iocall_next = NULL;
			context->last_exception = exc;
			context->run = WTRR_RAISE;
			return WTE_OK;
		}

		context->run = WTRR_RUN;

		iocall_next = context->iocall_next;

		if(NULL!=iocall_next)
		{
			context->iocall_next = NULL;
			context->run = iocall_next(context);
		}

		return WTE_OK;
	}
	else
	if(WTRR_EXTCALL==context->run)
	{
		const wt_u32* const ret_types = context->extcallinfo.retv;

		if(retc!=context->extcallinfo.retc || (0<retc && NULL==retv))
		{
			return WTE_BADARG;
		}

		for(i=0;i<retc;++i)
		{
			if(retv[i].type!=ret_types[i])
			{
				return WTE_BADARG;
			}
		}

		if(0!=(exc=wtr_ctx_push_values(context, retc, retv)))
		{
			context->last_exception = exc;
			context->run = WTRR_RAISE;
			return WTE_OK;
		}

		context->run = WTRR_RUN;
		return WTE_OK;
	}

	return WTE_BADCTX;
}

static void wtr_ctx_setup(struct wtr_context* const ctx, wt_u32 pc,
	struct wtr_program* const prog, struct wtr_function* const func,
	WTR_PFN_RETURNACTION return_action)
{
	ctx->last_exception = -1;
	ctx->frame.pc       = pc;
	ctx->frame.flags    = WTRCF_START;
	ctx->frame.ehandler = WTR_NULLPC;
	ctx->frame.dvt      = 0;
	ctx->frame.csos     = 0;
	ctx->frame.csos_sp  = 0;
	ctx->frame.tos_mark = 0;
	ctx->frame.dsp_mark = NULL;
	ctx->frame.prog     = prog;

	if(NULL!=prog)
	{
		++prog->refcount;
		ctx->called_prog = prog;
	}
	else
	{
		ctx->called_prog = NULL;
	}

	if(NULL!=func)
	{
		++func->refcount;
		ctx->called_func = func;
	}
	else
	{
		ctx->called_func = NULL;
	}

	ctx->return_action = return_action;
}

static wt_error wtr_ctx_invoke_prog(struct wtr_context* const ctx,
	struct wtr_program* const prog, wt_u32 pc,
	WTR_PFN_RETURNACTION return_action)
{
	if(WTRR_INIT!=ctx->run)
	{
		return WTE_BADCTX;
	}

	wtr_ctx_setup(ctx, pc, prog, NULL, return_action);

	ctx->retc = 0;
	ctx->run = WTRR_RUN;
	return WTE_OK;
}

static void ctor_return_action(struct wtr_context* const ctx)
{
	decrement_called_refcounts(ctx);
}

wt_error wtr_ctx_invoke_ctor(struct wtr_context* const ctx,
	struct wtr_program* const prog)
{
	return wtr_ctx_invoke_prog(ctx, prog, prog->pc_base+prog->pc_ctor,
		ctor_return_action);
}

static void dtor_return_action(struct wtr_context* const ctx)
{
	struct wtr_program* const prog = ctx->called_prog;

	decrement_called_refcounts(ctx);

	if(0==prog->refcount)
	{
		wtr_program_destroy(ctx->profile, prog);
	}
}

wt_error wtr_ctx_invoke_dtor(struct wtr_context* const ctx,
	struct wtr_program* const prog)
{
	return wtr_ctx_invoke_prog(ctx, prog, prog->pc_base+prog->pc_dtor,
		dtor_return_action);
}

wt_bool wtr_ctx_resize_value_array(struct wtr_context* const ctx,
	wt_u32 size)
{
	// TODO: if new size is not *that* smaller, then better to do nothing
	// (maybe do it in multiple of 8?)

	if(NULL==ctx->value_array)
	{
		if(0==size)
		{
			return WT_TRUE;
		}

		if(NULL==(ctx->value_array=(wt_value*)wtr_profile_sys_realloc(
			ctx->profile, NULL, 0, sizeof(wt_value)*size)))
		{
			return WT_FALSE;
		}

		ctx->value_arr_size = size;
	}
	else
	{
		wt_value* const new_arr = (wt_value*)wtr_profile_sys_realloc(
			ctx->profile, ctx->value_array,
			sizeof(wt_value)*ctx->value_arr_size, sizeof(wt_value)*size);

		if(NULL==new_arr && 0<size)
		{
			return WT_FALSE;
		}

		ctx->value_arr_size = size;
		ctx->value_array    = new_arr;
	}

	return WT_TRUE;
}

wt_exception wtr_ctx_pop_values(struct wtr_context* const ctx,
	wt_u32 nvalues, wt_value* const values, const wt_u32* types)
{
	wt_value*       arg = values;
	wt_value* const end = values + nvalues;

	struct wtr_program* const prog = ctx->frame.prog;

	/* number of needed number of octets on the DataStack */
	wt_u32 need = 0;

	if(0==nvalues)
	{
		return 0;
	}

	for(;arg<end;++types,++arg)
	{
		const wt_u32 sz = wtr_sizeof(*types);

		if(0==sz)
		{
			/* invalid type info */
			return BadProgramAccessException;
		}

		need += sz;

		arg->writable_type = *types;
		arg->v_u64         = 0;
	}

	if(WTR_CTX_DS_SIZE(ctx)<need)
	{
		return StackOverflowException;
	}

	for(--arg;values<=arg;--arg)
	{
		switch(arg->writable_type)
		{
		case WTT_CHAR:
		case WTT_BOOL:
		case WTT_U32:
		case WTT_I32:
#if WTRUNTIME_ENABLE_FLOAT32
		case WTT_F32:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
			{
				arg->v_u32 = *(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);
			}
			break;

		case WTT_U64:
		case WTT_I64:
#if WTRUNTIME_ENABLE_FLOAT64
		case WTT_F64:
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
			{
				arg->v_u64 = *(wt_u64*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u64);
			}
			break;

		case WTT_HANDLE:
			{
				arg->v_handle = *(wt_handle*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_handle);
			}
			break;

		case WTT_I8:
		case WTT_U8:
			{
				arg->v_u8 = (wt_u8)*(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);
			}
			break;

		case WTT_I16:
		case WTT_U16:
			{
				arg->v_u16 = (wt_u16)*(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);
			}
			break;

		case WTT_FILE:
		case WTT_DIR:
			{
				arg->v_file = (void*)(wt_uintptr)(*(wt_handle*)ctx->dsp);
				WTR_CTX_DS_POP(ctx, wt_handle);
			}
			break;

		case WTT_STRING:
			{
				arg->v_string = ((struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp))->data;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
			}
			break;

		case WTTF_SEQ|WTT_I8:
		case WTTF_SEQ|WTT_U8:
		case WTTF_SEQ|WTT_I16:
		case WTTF_SEQ|WTT_U16:
		case WTTF_SEQ|WTT_I32:
		case WTTF_SEQ|WTT_U32:
		case WTTF_SEQ|WTT_I64:
		case WTTF_SEQ|WTT_U64:
#if WTRUNTIME_ENABLE_FLOAT32
		case WTTF_SEQ|WTT_F32:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
		case WTTF_SEQ|WTT_F64:
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
		case WTTF_SEQ|WTT_CHAR:
		case WTTF_SEQ|WTT_BOOL:
		case WTTF_SEQ|WTT_HANDLE:
			{
				struct wtr_seq* const seq = (struct wtr_seq*)WTR_OBJ_FROM_VMADDR(prog, *(wt_vmptr*)ctx->dsp);
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				arg->v_seq.n = seq->length;
				arg->v_seq.p = seq->data;
			}
			break;

		default:
			return BadProgramAccessException;
		}
	}

	return 0;
}

wt_exception wtr_ctx_push_values(struct wtr_context* const ctx,
	const wt_u32 nvalues, wt_value* const values)
{
	wt_value*       val = values;
	wt_value* const end = values + nvalues;

	struct wtr_program* const prog = ctx->frame.prog;
	struct wtr_profile* const prof = ctx->profile;

	for(;val<end;++val)
	{
		const wt_u32 type_id = val->type;

		if(WTR_CTX_DS_SPACE(ctx)<wtr_sizeof(type_id))
		{
			return StackOverflowException;
		}

		switch(type_id)
		{
		case WTT_BOOL:
			WTR_CTX_DS_PUSH(ctx, wt_bool, val->v_bool);
			break;

		case WTT_I32:
		case WTT_U32:
#if WTRUNTIME_ENABLE_FLOAT32
		case WTT_F32:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
			WTR_CTX_DS_PUSH(ctx, wt_u32, val->v_u32);
			break;

		case WTT_I64:
		case WTT_U64:
#if WTRUNTIME_ENABLE_FLOAT64
		case WTT_F64:
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
			WTR_CTX_DS_PUSH(ctx, wt_u64, val->v_u64);
			break;

		case WTT_CHAR:
			WTR_CTX_DS_PUSH(ctx, wt_char, val->v_char);
			break;

		case WTT_HANDLE:
			WTR_CTX_DS_PUSH(ctx, wt_handle, val->v_handle);
			break;

		case WTT_I8:
		case WTT_U8:
			WTR_CTX_DS_PUSH(ctx, wt_u32, val->v_u8);
			break;

		case WTT_I16:
		case WTT_U16:
			WTR_CTX_DS_PUSH(ctx, wt_u32, val->v_u16);
			break;

		case WTT_FILE:
		case WTT_DIR:
			WTR_CTX_DS_PUSH(ctx, wt_handle, (wt_handle)((wt_uintptr)val->v_file));
			break;

		case WTT_STRING:
			{
				struct wtr_str* const sh = wtr_str_create(prog, val->v_string, 0, WT_TRUE);

				if(NULL!=sh)
				{
					const wt_vmptr      vm_addr = WTR_OBJ_TO_VMADDR(prog, sh);
					struct wtr_objentry ent;

					ent.pfn.ptr = prog;
					ent.pfn.ext = prof->pc_strd_leave;
					ent.obj     = vm_addr;

					if(wtr_dynstack_push(&ctx->tos, &ent))
					{
						WTR_CTX_DS_PUSH(ctx, wt_vmptr, vm_addr);
					}
					else
					{
						wtr_str_delete(prog, sh);
						return StackOverflowException;
					}
				}
				else
				{
					return OutOfMemoryException;
				}
			}
			break;

		default:
			return BadProgramAccessException;
		}
	}

	return 0;
}
