#include "config_private.h"
#include "processor.h"
#include "types.h"
#include "profile.h"
#include "program.h"
#include "ctx.h"
#include "stdlib.h"
#include "microprog.h"
#include "objects.h"
#include "function.h"
#include "funcsig.h"
#include "debug.h"
#include <string.h>
#if WTRUNTIME_DEBUG
#  include <assert.h>
#  include <stdio.h>
#endif /* WTRUNTIME_DEBUG */

#if WTRUNTIME_DEBUG
#  define goto__out_of_memory_exception() do { trap_out_of_memory(); goto out_of_memory_exception; } while(0)
void trap_out_of_memory() { }
#else /* WTRUNTIME_DEBUG */
#  define goto__out_of_memory_exception() goto out_of_memory_exception
#endif /* WTRUNTIME_DEBUG */

#define CHECK_PUSH(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SPACE(_ctx_)>=(_size_)) \
		{ \
			/* skip */ \
		} \
		else \
		{ \
			goto stack_overflow_exception; \
		} \
	} \
	while(0)

#define CHECK_POP(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SIZE(_ctx_)>=(_size_)) \
		{ \
			/* skip */ \
		} \
		else \
		{ \
			goto stack_underflow_exception; \
		} \
	} \
	while(0)

#define READ_LONGIMM \
	do \
	{ \
		opdata = (opdata << WT_TOMB_INSTRBITS) | (code[ctx->frame.pc++]); \
	} \
	while(0)

#define READ_STACKIMM \
	do \
	{ \
		opdata = *(wt_u32*)ctx->dsp; \
		ctx->dsp += sizeof(wt_u32); \
	} \
	while(0)

#ifndef RotateRight8
#  define RotateRight8(value,shift) (((value)>>(shift)) | ((value)<<(8-(shift))))
#endif /* RotateRight8 */

#ifndef RotateRight16
#  define RotateRight16(value,shift) (((value)>>(shift)) | ((value)<<(16-(shift))))
#endif /* RotateRight16 */

#ifndef RotateRight32
#  define RotateRight32(value,shift) (((value)>>(shift)) | ((value)<<(32-(shift))))
#endif /* RotateRight32 */

#ifndef RotateRight64
#  define RotateRight64(value,shift) (((value)>>(shift)) | ((value)<<(64-(shift))))
#endif /* RotateRight64 */

#define __OP_FUNC_ADD(_lhs_,_rhs_)    ((_lhs_) + (_rhs_))
#define __OP_FUNC_SUB(_lhs_,_rhs_)    ((_lhs_) - (_rhs_))
#define __OP_FUNC_MUL(_lhs_,_rhs_)    ((_lhs_) * (_rhs_))
#define __OP_FUNC_DIV(_lhs_,_rhs_)    ((_lhs_) / (_rhs_))
#define __OP_FUNC_SMUL(_lhs_,_rhs_)   ((_lhs_) * (_rhs_))
#define __OP_FUNC_SDIV(_lhs_,_rhs_)   ((_lhs_) / (_rhs_))
#define __OP_FUNC_SHL(_lhs_,_rhs_)    ((_lhs_) << (_rhs_))
#define __OP_FUNC_SHR(_lhs_,_rhs_)    ((_lhs_) >> (_rhs_))
#define __OP_FUNC_SAR(_lhs_,_rhs_)    ((_lhs_) >> (_rhs_))
#define __OP_FUNC_ROR(_lhs_,_rhs_)    (RotateRight32(_lhs_,_rhs_))
#define __OP_FUNC_MOD(_lhs_,_rhs_)    ((_lhs_) % (_rhs_))
#define __OP_FUNC_SMOD(_lhs_,_rhs_)   ((_lhs_) % (_rhs_))
#define __OP_FUNC_LSOR(_lhs_,_rhs_)   (((_lhs_) << 16) | (_rhs_))
#define __OP_FUNC_RSB(_lhs_,_rhs_)    ((_rhs_) - (_lhs_))
#define __OP_FUNC_ADDL(_lhs_,_rhs_)   ((_lhs_) + (_rhs_))
#define __OP_FUNC_SUBL(_lhs_,_rhs_)   ((_lhs_) - (_rhs_))
#define __OP_FUNC_MULL(_lhs_,_rhs_)   ((_lhs_) * (_rhs_))
#define __OP_FUNC_DIVL(_lhs_,_rhs_)   ((_lhs_) / (_rhs_))
#define __OP_FUNC_SMULL(_lhs_,_rhs_)  ((_lhs_) * (_rhs_))
#define __OP_FUNC_SDIVL(_lhs_,_rhs_)  ((_lhs_) / (_rhs_))
#define __OP_FUNC_SHLL(_lhs_,_rhs_)   ((_lhs_) << (_rhs_))
#define __OP_FUNC_SHRL(_lhs_,_rhs_)   ((_lhs_) >> (_rhs_))
#define __OP_FUNC_SARL(_lhs_,_rhs_)   ((_lhs_) >> (_rhs_))
#define __OP_FUNC_RORL(_lhs_,_rhs_)   (RotateRight64(_lhs_,_rhs_))
#define __OP_FUNC_XOR(_lhs_,_rhs_)    ((_lhs_) ^ (_rhs_))
#define __OP_FUNC_BAND(_lhs_,_rhs_)   ((_lhs_) & (_rhs_))
#define __OP_FUNC_BOR(_lhs_,_rhs_)    ((_lhs_) | (_rhs_))
#define __OP_FUNC_XORL(_lhs_,_rhs_)   ((_lhs_) ^ (_rhs_))
#define __OP_FUNC_BANDL(_lhs_,_rhs_)  ((_lhs_) & (_rhs_))
#define __OP_FUNC_BORL(_lhs_,_rhs_)   ((_lhs_) | (_rhs_))
#define __OP_FUNC_LSORL(_lhs_,_rhs_)  (((_lhs_) << 24) | (_rhs_))
#define __OP_FUNC_MODL(_lhs_,_rhs_)   ((_lhs_) % (_rhs_))
#define __OP_FUNC_SMODL(_lhs_,_rhs_)  ((_lhs_) % (_rhs_))
#define __OP_FUNC_FADD(_lhs_,_rhs_)   ((_lhs_) + (_rhs_))
#define __OP_FUNC_FSUB(_lhs_,_rhs_)   ((_lhs_) - (_rhs_))
#define __OP_FUNC_FMUL(_lhs_,_rhs_)   ((_lhs_) * (_rhs_))
#define __OP_FUNC_FDIV(_lhs_,_rhs_)   ((_lhs_) / (_rhs_))
#define __OP_FUNC_FADDL(_lhs_,_rhs_)  ((_lhs_) + (_rhs_))
#define __OP_FUNC_FSUBL(_lhs_,_rhs_)  ((_lhs_) - (_rhs_))
#define __OP_FUNC_FMULL(_lhs_,_rhs_)  ((_lhs_) * (_rhs_))
#define __OP_FUNC_FDIVL(_lhs_,_rhs_)  ((_lhs_) / (_rhs_))
#if WTRUNTIME_HAVE_F_MATH
#define __OP_FUNC_FMOD(_lhs_,_rhs_)   (fmodf((_lhs_), (_rhs_)))
#else /* WTRUNTIME_HAVE_F_MATH */
#define __OP_FUNC_FMOD(_lhs_,_rhs_)   ((wt_f32)fmod((_lhs_), (_rhs_)))
#endif /* WTRUNTIME_HAVE_F_MATH */
#define __OP_FUNC_FMODL(_lhs_,_rhs_)  (fmod((_lhs_), (_rhs_)))
#define __OP_FUNC_LAND(_lhs_,_rhs_)   ((_lhs_) && (_rhs_))
#define __OP_FUNC_LOR(_lhs_,_rhs_)    ((_lhs_) || (_rhs_))
#define __OP_FUNC_LXOR(_lhs_,_rhs_)   ((_lhs_) ^ (_rhs_))
#define __OP_FUNC_ADDB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) + ((wt_u8)(_rhs_))))
#define __OP_FUNC_SUBB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) - ((wt_u8)(_rhs_))))
#define __OP_FUNC_MULB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) * ((wt_u8)(_rhs_))))
#define __OP_FUNC_SMULB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) * ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_DIVB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) / ((wt_u8)(_rhs_))))
#define __OP_FUNC_SDIVB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) / ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_MODB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) % ((wt_u8)(_rhs_))))
#define __OP_FUNC_SMODB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) % ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_SHLB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) << ((wt_u8)(_rhs_))))
#define __OP_FUNC_RORB(_lhs_,_rhs_)   ((wt_u8)(RotateRight8((wt_u8)(_lhs_), (wt_u8)(_rhs_))))
#define __OP_FUNC_ADDH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) + ((wt_u16)(_rhs_))))
#define __OP_FUNC_SUBH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) - ((wt_u16)(_rhs_))))
#define __OP_FUNC_MULH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) * ((wt_u16)(_rhs_))))
#define __OP_FUNC_SMULH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) * ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_DIVH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) / ((wt_u16)(_rhs_))))
#define __OP_FUNC_SDIVH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) / ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_MODH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) % ((wt_u16)(_rhs_))))
#define __OP_FUNC_SMODH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) % ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_SHLH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) << ((wt_u16)(_rhs_))))
#define __OP_FUNC_RORH(_lhs_,_rhs_)   ((wt_u16)(RotateRight16((wt_u16)(_lhs_), (wt_u16)(_rhs_))))

#define __OP_CMP_CMP(_lhs_,_rhs_)     (((_lhs_)<(_rhs_))?(-1):(((_lhs_)>(_rhs_)?1:0)))
#define __OP_CMP_SCMP(_lhs_,_rhs_)    __OP_CMP_CMP(_lhs_, _rhs_)
#define __OP_CMP_CMPL(_lhs_,_rhs_)    __OP_CMP_CMP(_lhs_, _rhs_)
#define __OP_CMP_SCMPL(_lhs_,_rhs_)   __OP_CMP_CMP(_lhs_, _rhs_)
#define __OP_CMP_FCMP(_lhs_,_rhs_)    __OP_CMP_CMP(_lhs_, _rhs_)
#define __OP_CMP_FCMPL(_lhs_,_rhs_)   __OP_CMP_CMP(_lhs_, _rhs_)
#define __OP_CMP_SCMPB(_lhs_,_rhs_)   __OP_CMP_CMP(((wt_i8)(wt_u8)(_lhs_)), ((wt_i8)(wt_u8)(_rhs_)))
#define __OP_CMP_SCMPH(_lhs_,_rhs_)   __OP_CMP_CMP(((wt_i16)(wt_u16)(_lhs_)), ((wt_i16)(wt_u16)(_rhs_)))

#define __OP_CHECK_NOTHING(_rhs_)
#define __OP_CHECK_DIV_NONZERO(_rhs_) if(!(_rhs_)) { goto division_by_zero_exception; }

#define _OP_I_S(_opcode_,_operand_type_,_check_) \
\
		case WT_TOMB_OP_##_opcode_##_I: /* Immediate */ \
			{ \
				CHECK_POP(ctx, sizeof(_operand_type_)); \
				__OP_CHECK_##_check_ (opdata); \
				*(_operand_type_*)ctx->dsp = \
					__OP_FUNC_##_opcode_(*(_operand_type_*)ctx->dsp, opdata); \
			} \
			break; \
\
		case WT_TOMB_OP_##_opcode_##_S: /* StackImmediate */ \
			{ \
				READ_STACKIMM; \
				CHECK_POP(ctx, sizeof(_operand_type_)); \
				__OP_CHECK_##_check_ (opdata); \
				*(_operand_type_*)ctx->dsp = \
					__OP_FUNC_##_opcode_(*(_operand_type_*)ctx->dsp, opdata); \
			} \
			break; \

#define _OP_I_L(_opcode_,_operand_type_,_check_) \
\
		case WT_TOMB_OP_##_opcode_##_I: /* Immediate */ \
			{ \
				CHECK_POP(ctx, sizeof(_operand_type_)); \
				__OP_CHECK_##_check_ (opdata); \
				*(_operand_type_*)ctx->dsp = \
					__OP_FUNC_##_opcode_(*(_operand_type_*)ctx->dsp, opdata); \
			} \
			break; \
\
		case WT_TOMB_OP_##_opcode_##_L: /* LongImmediate */ \
			{ \
				READ_LONGIMM; \
				CHECK_POP(ctx, sizeof(_operand_type_)); \
				__OP_CHECK_##_check_ (opdata); \
				*(_operand_type_*)ctx->dsp = \
					__OP_FUNC_##_opcode_(*(_operand_type_*)ctx->dsp, opdata); \
			} \
			break; \

#define _OP_(_opcode_,_operand_lhs_type_,_operand_rhs_type_,_check_) \
\
		case WT_TOMB_OP_##_opcode_: /* custom */ \
			{ \
				_operand_rhs_type_ rhs; \
				CHECK_POP(ctx, sizeof(_operand_lhs_type_)+sizeof(_operand_rhs_type_)); \
				rhs = *(_operand_rhs_type_*)ctx->dsp; \
				WTR_CTX_DS_POP(ctx, _operand_rhs_type_); \
				__OP_CHECK_##_check_ (rhs); \
				*(_operand_lhs_type_*)ctx->dsp = \
					__OP_FUNC_##_opcode_(*(_operand_lhs_type_*)ctx->dsp, rhs); \
			} \
			break; \

#define _OP_cmp(_opcode_,_operand_type_) \
\
		case WT_TOMB_OP_##_opcode_: \
			{ \
				_operand_type_ rhs; \
				_operand_type_ lhs; \
				CHECK_POP(ctx, sizeof(_operand_type_)*2); \
				rhs = *(_operand_type_*)ctx->dsp; \
				WTR_CTX_DS_POP(ctx, _operand_type_); \
				lhs = *(_operand_type_*)ctx->dsp; \
				WTR_CTX_DS_POP(ctx, _operand_type_); \
				WTR_CTX_DS_PUSH(ctx, wt_i32, \
					__OP_CMP_##_opcode_(lhs, rhs)); \
			} \
			break; \

static enum wtr_runresult wtr_processor_op_OSF(struct wtr_profile* const prof, struct wtr_context* const ctx)
{
	struct wtr_program* const prog = ctx->frame.prog;

	((void)prof);

	if(WTRCF_CSOS & ctx->frame.flags)
	{
		struct wtr_objentry* csos = ctx->frame.csos;
		struct wtr_objentry* ent  = csos + ctx->frame.csos_sp;
		wt_u32               cnt;
		wt_u32               size;

		while(ent>csos)
		{
			struct wtr_dynstack* stk = &ctx->rtos;

			/* the 'ent' points right after the last real element */

			--ent;

			/* the 'ent.obj' is a pointer into the DVT */

			ent->obj = *(wt_vmptr*)WTR_PTR_FROM_VMADDR(prog, ent->obj);

			/* add entry into TOS or RTOS */
			/* (add to TOS only if it is empty) */

			if(ctx->tos.sp!=ctx->frame.tos_mark)
			{
				/* TOS is unlikely empty, even if the CSOS has more than one element */
				/* take this empty branch as a branch-prediction optimization */
			}
			else
			{
				stk = &ctx->tos;
			}

#ifdef WTRUNTIME_DEBUG
			{
			const wt_bool dynstack_push_succeeded =
#endif /* WTRUNTIME_DEBUG */
			wtr_dynstack_push(stk, ent); /* TODO: if the push fails, then it cause a memory leak */
			/* solution: emergency space and raise an exception */

#ifdef WTRUNTIME_DEBUG
			assert(dynstack_push_succeeded);
			}
#endif /* WTRUNTIME_DEBUG */
		}

		cnt  = *(wt_u32*)ctx->csp;
		size = cnt * sizeof(struct wtr_objentry);

		WTR_CTX_CS_POP(ctx, wt_u32);

		ctx->csp          += size;
		ctx->frame.flags  &= ~WTRCF_CSOS;
		ctx->frame.csos    = NULL;
		ctx->frame.csos_sp = 0;

		WTR_CTX_CS_POP(ctx, wt_u32);
	}

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_DVTF(struct wtr_profile* const prof, struct wtr_context* const ctx)
{
	((void)prof);

	if(0==(WTRCF_CSOS & ctx->frame.flags))
	{
		/* this is the more likely branch */
	}
	else
	{
		ctx->last_exception = BadProgramAccessException;
		return WTRR_RAISE;
	}

	if(WTRCF_DVT & ctx->frame.flags)
	{
		wt_u32 dvt_size = *(wt_u32*)ctx->csp;
		WTR_CTX_CS_POP(ctx, wt_u32);

		ctx->csp         += dvt_size;
		ctx->frame.flags &= ~WTRCF_DVT;
		ctx->frame.dvt    = 0;
	}

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_LEAVE(struct wtr_profile* const prof, struct wtr_context* const ctx)
{
	if(WTRCF_CSOS & ctx->frame.flags)
	{
		// TODO: handle return value?
		wtr_processor_op_OSF(prof, ctx);
	}

	if(WTRCF_DVT & ctx->frame.flags)
	{
		// TODO: handle return value?
		wtr_processor_op_DVTF(prof, ctx);
	}

	if(WTRCF_START & ctx->frame.flags)
	{
		WTR_PFN_RETURNACTION return_action = ctx->return_action;

		if(NULL!=return_action)
		{
			ctx->return_action = NULL;
			return_action(ctx);
		}

		return WTRR_RETURNED;
	}

	memcpy(&ctx->frame, ctx->csp, sizeof(ctx->frame));
	ctx->csp += sizeof(ctx->frame);

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_unwind_frames(
	struct wtr_profile* const prof, struct wtr_context* const ctx)
{
	enum wtr_runresult rr = WTRR_RUN;

	ctx->last_exception = *(wt_exception*)ctx->dsp;

	while(WTRR_RUN==rr && WTR_NULLPC==ctx->frame.ehandler)
	{
		rr = wtr_processor_op_LEAVE(prof, ctx);
	}

	if(WTRR_RETURNED==rr)
	{
		return WTRR_EXCEPTION;
	}

	if(WTRR_RUN!=rr)
	{
		return rr;
	}

	ctx->dsp            = ctx->frame.dsp_mark;
	ctx->frame.pc       = ctx->frame.ehandler;
	ctx->frame.ehandler = WTR_NULLPC;

	WTR_CTX_DS_PUSH(ctx, wt_exception, ctx->last_exception);
	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_CFP(struct wtr_profile* const prof, struct wtr_context* const ctx, wt_pfn* pfn)
{
	if(pfn->ext==prof->external_tag)
	{
		struct wtr_cfunction* const cfunc = (struct wtr_cfunction*)pfn->ptr;
		const wt_u32                argc  = cfunc->funcsig->argc;

		if(!wtr_ctx_resize_value_array(ctx, argc))
		{
			ctx->last_exception = OutOfMemoryException;
			return WTRR_RAISE;
		}

		/* build the extcallinfo */

		ctx->extcallinfo.prof  = prof;
		ctx->extcallinfo.func  = cfunc->extfunc_pfn;
		ctx->extcallinfo.token = cfunc->token;
		ctx->extcallinfo.argc  = argc;
		ctx->extcallinfo.retc  = cfunc->funcsig->retc;
		ctx->extcallinfo.retv  = cfunc->funcsig->retv;

		/* pop function arguments */

		if(0!=(ctx->last_exception=wtr_ctx_pop_values(ctx, argc,
			ctx->value_array, cfunc->funcsig->argv)))
		{
			return WTRR_RAISE;
		}

		ctx->run = WTRR_EXTCALL;

		// TODO: only if context allows it
		if(1)
		{
			ctx->extcallinfo.func(ctx, &ctx->extcallinfo, ctx->value_array);
		}

		return ctx->run;
	}
	else
	{
		/* is there enough space to store the current frame? */
		if(WTR_CTX_CS_SPACE(ctx)>=sizeof(ctx->frame))
		{
			/* store the current frame */
			WTR_CTX_CS_PUSH(ctx, struct wtr_callframe, ctx->frame);

			/* assemble new frame */
			ctx->frame.pc       = pfn->ext;
			ctx->frame.flags    = 0;
			ctx->frame.ehandler = WTR_NULLPC;
			ctx->frame.dvt      = 0;
			ctx->frame.csos     = 0;
			ctx->frame.csos_sp  = 0;
			ctx->frame.tos_mark = ctx->tos.sp;
			ctx->frame.dsp_mark = NULL;
			ctx->frame.prog     = (struct wtr_program*)pfn->ptr;
		}
		else
		{
			ctx->last_exception = StackOverflowException;
			return WTRR_RAISE;
		}
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_processor_tomb_run(struct wtr_profile* const prof,
	struct wtr_context* const ctx, wt_u32 quantum)
{
	/* the actual code memory space */
	const wt_u16* const code = (const wt_u16*)prof->memory;

	/* largest PC: it cannot exceed the allocated memory */
	const wt_u32 maxpc = (prof->total_memsize / sizeof(*code)) - 1;

	/* exception ID buffer for 'raise_exception' */
	wt_exception exc;

	if(WTRR_RAISE==ctx->run)
	{
		exc = ctx->last_exception;
		goto raise_exception;
	}

	while(WTRR_RUN==ctx->run)
	{
		wt_u16 instr;
		wt_u32 opdata;

#ifdef WTRUNTIME_DEBUG
		wtr_tomb_instructionset opcode;
#endif /* WTRUNTIME_DEBUG */

		if(0==(quantum--))
		{
			ctx->run = WTRR_QUANTUM;
			return WTRR_QUANTUM;
		}

		if(ctx->frame.pc>=maxpc)
		{
			goto access_violation_exception;
		}

		instr  = code[ctx->frame.pc++];
		opdata = WT_TOMB_GETOPDATA(instr);

#ifdef WTRUNTIME_DEBUG
		opcode = (wtr_tomb_instructionset)WT_TOMB_GETOPCODE(instr);
		opcode = opcode;
#endif /* WTRUNTIME_DEBUG */

#ifdef DISPLAY_EXECUTED_INSTRUCTIONS
		printf("%08X/%08X %02X %02X %04X\n", (ctx->frame.pc-ctx->frame.prog->pc_base-1)*2,
			(ctx->frame.pc-ctx->frame.prog->pc_base-1)*2, WT_TOMB_GETOPCODE(instr), opdata, code[ctx->frame.pc]);
#endif /* DISPLAY_EXECUTED_INSTRUCTIONS */

#ifdef DISPLAY_DATASTACK_AFTER_INSTRUCTIONS
		{
			wt_u32  i = 0;
			wt_u32* p = (wt_u32*)ctx->dsp;

			for(i=0;p<(wt_u32*)ctx->ds_hi;++i)
			{
				if(0==(i%8))
				{
					if(i)
					{
						printf("\n");
					}

					printf("   ");
				}

				printf(" %08X", *p++);
			}

			printf("\n\n");
		}
#endif /* DISPLAY_DATASTACK_AFTER_INSTRUCTIONS */

		switch(WT_TOMB_GETOPCODE(instr))
		{
		case WT_TOMB_OP_NOP:
			{
				/* no operation */
			}
			break;

		case WT_TOMB_OP_LD_I: /* Load Immediate */
			{
				wt_vmptr addr;
__ld_i:
				CHECK_POP(ctx, sizeof(addr));
				addr = *(wt_vmptr*)ctx->dsp;
				if((opdata & 0x3)==0)
				{
					if(WTR_CTX_DS_SPACE(ctx)>=opdata)
					{
						void* const mem = WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr);
						ctx->dsp -= (opdata - sizeof(addr));
						memcpy(ctx->dsp, mem, opdata);
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_LD_L: /* Load LongImmediate */
			{
				READ_LONGIMM;
				goto __ld_i;
			}
			break;

		case WT_TOMB_OP_LD_S: /* Load StackImmediate */
			{
				READ_STACKIMM;
				goto __ld_i;
			}
			break;

		case WT_TOMB_OP_ST_I: /* Store Immediate */
			{
				wt_vmptr addr;
__st_i:
				CHECK_POP(ctx, sizeof(addr)+opdata);
				addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				if(0==(0x3 & opdata))
				{
					void* const ptr = WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr);
					memcpy(ptr, ctx->dsp, opdata);
					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_ST_L: /* Store LongImmediate */
			{
				READ_LONGIMM;
				goto __st_i;
			}
			break;

		case WT_TOMB_OP_ST_S: /* Store StackImmediate */
			{
				READ_STACKIMM;
				goto __st_i;
			}
			break;

		case WT_TOMB_OP_IDRD_I: /* Internal DataStack Dereference Immediate */
			{
__idrd_i:
				if(0==(0x3 & opdata) && WTR_CTX_DS_SIZE(ctx)>=opdata)
				{
					wt_vmptr* const addr = (wt_vmptr*)(ctx->dsp + opdata);
					*addr = *(wt_vmptr*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, *addr);
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_IDRD_L: /* Internal DataStack Dereference LongImmediate */
			{
				READ_LONGIMM;
				goto __idrd_i;
			}
			break;

		case WT_TOMB_OP_IDRD_S: /* Internal DataStack Dereference StackImmediate */
			{
				READ_STACKIMM;
				goto __idrd_i;
			}
			break;

		case WT_TOMB_OP_PBA_I: /* Push Base Address Immediate */
			{
				wt_vmptr addr = opdata + ctx->frame.prog->base_address;
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
			}
			break;

		case WT_TOMB_OP_PBA_L: /* Push Base Address LongImmediate */
			{
				wt_vmptr addr;
				READ_LONGIMM;
				addr = opdata + ctx->frame.prog->base_address;
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
			}
			break;

		case WT_TOMB_OP_PBA_S: /* Push Base Address StackImmediate */
			{
				wt_vmptr addr;
				READ_STACKIMM;
				addr = opdata + ctx->frame.prog->base_address;
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
			}
			break;

		case WT_TOMB_OP_PVT_I: /* Push Variable Table Immediate */
			{
__pvt_i:
				if(ctx->frame.flags & WTRCF_DVT)
				{
					wt_vmptr addr = opdata + ctx->frame.dvt;
					CHECK_PUSH(ctx, sizeof(wt_vmptr));
					WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_PVT_L: /* Push Variable Table LongImmediate */
			{
				READ_LONGIMM;
				goto __pvt_i;
			}
			break;

		case WT_TOMB_OP_PVT_S: /* Push Variable Table StackImmediate */
			{
				READ_STACKIMM;
				goto __pvt_i;
			}
			break;

		case WT_TOMB_OP_PPC: /* Push Program Counter */
			{
				CHECK_PUSH(ctx, sizeof(ctx->frame.pc));
				WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->frame.pc);
			}
			break;

		case WT_TOMB_OP_DUP_I: /* Duplicate Immediate */
			{
				void* prev_dsp;
				if((opdata & 0x3)==0)
				{
					CHECK_POP(ctx, opdata);
					CHECK_PUSH(ctx, opdata);
					prev_dsp = ctx->dsp;
					ctx->dsp -= opdata;
					memcpy(ctx->dsp, prev_dsp, opdata);
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_DUP_L: /* Duplicate LongImmediate */
			{
				void* prev_dsp;
				READ_LONGIMM;
				if((opdata & 0x3)==0)
				{
					CHECK_POP(ctx, opdata);
					CHECK_PUSH(ctx, opdata);
					prev_dsp = ctx->dsp;
					ctx->dsp -= opdata;
					memcpy(ctx->dsp, prev_dsp, opdata);
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_POP_I: /* Pop Immediate */
			{
				if((opdata & 0x3)==0)
				{
					CHECK_POP(ctx, opdata);
					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_POP_L: /* Pop LongImmediate */
			{
				READ_LONGIMM;
				if((opdata & 0x3)==0)
				{
					CHECK_POP(ctx, opdata);
					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_PWZ_I: /* Push Word Zero Immediate */
			{
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWZ_L: /* Push Word Zero LongImmediate */
			{
				READ_LONGIMM;
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWO_I: /* Push Word One Immediate */
			{
				opdata |= ~WT_TOMB_OPDATAMASK;
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWO_L: /* Push Word One LongImmediate */
			{
				READ_LONGIMM;
				opdata |= ~WT_TOMB_MASKLONGIMM;
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWZL_I: /* Push LongWord Zero Immediate */
			{
				wt_u64 longword = opdata;
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_PWZL_L: /* Push LongWord Zero LongImmediate */
			{
				wt_u64 longword;
				READ_LONGIMM;
				longword = opdata;
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_PWOL_I: /* Push LongWord One Immediate */
			{
				wt_u64 longword = opdata | (~(wt_u64)WT_TOMB_OPDATAMASK);
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_PWOL_L: /* Push LongWord One LongImmediate */
			{
				wt_u64 longword;
				READ_LONGIMM;
				longword = opdata | (~(wt_u64)WT_TOMB_MASKLONGIMM);
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_DVTA_I: /* Allocate DVT Immediate */
			{
__dvta_i:
				if((ctx->frame.flags & (WTRCF_DVT|WTRCF_CSOS))==0 && (opdata&0x3)==0)
				{
					if(WTR_CTX_CS_SPACE(ctx)>=sizeof(wt_u32)+opdata)
					{
						ctx->csp -= opdata;
						ctx->frame.dvt = WTR_PTR_TO_VMADDR(ctx->frame.prog, ctx->csp);
						memset(ctx->csp, 0, opdata); /* clear the DVT */
						WTR_CTX_CS_PUSH(ctx, wt_u32, opdata);
						ctx->frame.flags |= WTRCF_DVT;
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_DVTA_L: /* Allocate DVT LongImmediate */
			{
				READ_LONGIMM;
				goto __dvta_i;
			}
			break;

		case WT_TOMB_OP_DVTA_S: /* Allocate DVT StackImmediate */
			{
				READ_STACKIMM;
				goto __dvta_i;
			}
			break;

		case WT_TOMB_OP_DVTF: /* Free DVT */
			{
				ctx->run = wtr_processor_op_DVTF(prof, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_OSA_I: /* Allocate CSOS Immediate */
			{
__osa_i:
				if((ctx->frame.flags & (WTRCF_DVT|WTRCF_CSOS))==WTRCF_DVT)
				{
					if(WTR_CTX_CS_SPACE(ctx)>=sizeof(wt_u32)*2+opdata)
					{
						size_t size = opdata * sizeof(struct wtr_objentry);

						WTR_CTX_CS_PUSH(ctx, wt_u32, opdata);
						ctx->csp -= size;
						ctx->frame.csos = (struct wtr_objentry*)ctx->csp;
						memset(ctx->csp, 0, size); /* clear the CSOS */
						WTR_CTX_CS_PUSH(ctx, wt_u32, opdata);
						ctx->frame.flags |= WTRCF_CSOS;
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSA_L: /* Allocate CSOS LongImmediate */
			{
				READ_LONGIMM;
				goto __osa_i;
			}
			break;

		case WT_TOMB_OP_OSA_S: /* Allocate CSOS StackImmediate */
			{
				READ_STACKIMM;
				goto __osa_i;
			}
			break;

		case WT_TOMB_OP_OSF: /* Free CSOS */
			{
				ctx->run = wtr_processor_op_OSF(prof, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_OSPU: /* CSOS Push */
			{
				wt_u32 csos_size = *(((wt_u32*)ctx->frame.csos) - 1);

				if(WTRCF_CSOS & ctx->frame.flags)
				{
					if(ctx->frame.csos_sp<csos_size)
					{
						memcpy(ctx->frame.csos + (ctx->frame.csos_sp++), ctx->dsp, sizeof(struct wtr_objentry));
						ctx->dsp += sizeof(struct wtr_objentry);
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSPO: /* CSOS Pop */
			{
				if(ctx->frame.flags & WTRCF_CSOS)
				{
					if(0!=ctx->frame.csos_sp)
					{
						CHECK_PUSH(ctx, sizeof(struct wtr_objentry));

						ctx->dsp -= sizeof(struct wtr_objentry);
						memcpy(ctx->dsp, ctx->frame.csos + (--ctx->frame.csos_sp), sizeof(struct wtr_objentry));
					}
					else
					{
						goto stack_underflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSSP: /* CSOS SP */
			{
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->frame.csos_sp);
			}
			break;

		case WT_TOMB_OP_TSZ: /* TOS Size */
			{
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->tos.sp - ctx->frame.tos_mark);
			}
			break;

		case WT_TOMB_OP_TPU: /* TOS Push */
			{
				CHECK_POP(ctx, sizeof(struct wtr_objentry));
				if(wtr_dynstack_push(&ctx->tos, (struct wtr_objentry*)ctx->dsp))
				{
					WTR_CTX_DS_POP(ctx, struct wtr_objentry);
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TPO: /* TOS Pop */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					struct wtr_objentry* top = wtr_dynstack_top(&ctx->tos);

					CHECK_PUSH(ctx, sizeof(struct wtr_objentry));
					WTR_CTX_DS_PUSH(ctx, struct wtr_objentry, *top);
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
_put_one_from_rtos_to_tos:
						top = wtr_dynstack_top(&ctx->rtos);
						wtr_dynstack_push(&ctx->tos, top);
						wtr_dynstack_pop(&ctx->rtos);

						break;
					}
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TDR: /* TOS Drop */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
						goto _put_one_from_rtos_to_tos;
					}
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TCL: /* TOS Clear */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					CHECK_PUSH(ctx, sizeof(wt_u32));
					WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->frame.pc);
					ctx->frame.pc = prof->pc_tcl_ja;
				}
			}
			break;

		case WT_TOMB_OP_PDSP_I: /* Push Data Stack Pointer Immediate */
			{
				wt_vmptr dsp;
__pdsp_i:
				dsp = opdata + WTR_PTR_TO_VMADDR(ctx->frame.prog, ctx->dsp);
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, dsp);
			}
			break;

		case WT_TOMB_OP_PDSP_L: /* Push Data Stack Pointer LongImmediate */
			{
				READ_LONGIMM;
				goto __pdsp_i;
			}
			break;

		case WT_TOMB_OP_VAS_I: /* Virtual Address Size Immediate */
			{
__vas_i:
				opdata *= sizeof(wt_vmptr);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_VAS_L: /* Virtual Address Size LongImmediate */
			{
				READ_LONGIMM;
				goto __vas_i;
			}
			break;

		case WT_TOMB_OP_ENTER_I: /* Enter Immediate */
			{
__enter_i:
				/* is there enough space to store the current frame? */
				if(WTR_CTX_CS_SPACE(ctx)>=sizeof(ctx->frame))
				{
					/* store the current frame */
					WTR_CTX_CS_PUSH(ctx, struct wtr_callframe, ctx->frame);

					/* assemble new frame */
					ctx->frame.pc       = ctx->frame.prog->pc_base + opdata;
					ctx->frame.flags    = 0;
					ctx->frame.ehandler = WTR_NULLPC;
					ctx->frame.dvt      = 0;
					ctx->frame.csos     = 0;
					ctx->frame.csos_sp  = 0;
					ctx->frame.tos_mark = ctx->tos.sp;
					ctx->frame.dsp_mark = NULL;
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_ENTER_L: /* Enter LongImmediate */
			{
				READ_LONGIMM;
				goto __enter_i;
			}
			break;

		case WT_TOMB_OP_ENTER_S: /* Enter StackImmediate */
			{
				READ_STACKIMM;
				goto __enter_i;
			}
			break;

		case WT_TOMB_OP_LEAVE: /* Leave current frame */
			{
				ctx->run = wtr_processor_op_LEAVE(prof, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_JMP_I: /* Jump Immediate */
			{
				if(opdata & WT_TOMB_HIBITIMM)
				{
					opdata |= ~WT_TOMB_OPDATAMASK;
				}

				ctx->frame.pc += opdata - 1;
			}
			break;

		case WT_TOMB_OP_JMP_L: /* Jump LongImmediate */
			{
				READ_LONGIMM;
				if(opdata & WT_TOMB_HIBITLONGIMM)
				{
					opdata |= ~WT_TOMB_MASKLONGIMM;
				}

				ctx->frame.pc += opdata - 2;
			}
			break;

		case WT_TOMB_OP_JMP_S: /* Jump StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.pc += opdata - 1;
			}
			break;

		case WT_TOMB_OP_JT_I: /* JumpTrue Immediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				if(cond)
				{
					if(opdata & WT_TOMB_HIBITIMM)
					{
						opdata |= ~WT_TOMB_OPDATAMASK;
					}

					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JT_L: /* JumpTrue LongImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				READ_LONGIMM;
				if(cond)
				{
					if(opdata & WT_TOMB_HIBITLONGIMM)
					{
						opdata |= ~WT_TOMB_MASKLONGIMM;
					}

					ctx->frame.pc += opdata - 2;
				}
			}
			break;

		case WT_TOMB_OP_JT_S: /* JumpTrue StackImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				READ_STACKIMM;
				if(cond)
				{
					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JF_I: /* JumpFalse Immediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				if(!cond)
				{
					if(opdata & WT_TOMB_HIBITIMM)
					{
						opdata |= ~WT_TOMB_OPDATAMASK;
					}

					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JF_L: /* JumpFalse LongImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				READ_LONGIMM;
				if(!cond)
				{
					if(opdata & WT_TOMB_HIBITLONGIMM)
					{
						opdata |= ~WT_TOMB_MASKLONGIMM;
					}

					ctx->frame.pc += opdata - 2;
				}
			}
			break;

		case WT_TOMB_OP_JF_S: /* JumpFalse StackImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				cond = *(wt_bool*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_bool);

				READ_STACKIMM;
				if(!cond)
				{
					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JA_I: /* Jump Absolute Immediate */
			{
				ctx->frame.pc = ctx->frame.prog->pc_base + opdata;
			}
			break;

		case WT_TOMB_OP_JA_L: /* Jump Absolute LongImmediate */
			{
				READ_LONGIMM;
				ctx->frame.pc = ctx->frame.prog->pc_base + opdata;
			}
			break;

		case WT_TOMB_OP_JA_S: /* Jump Absolute StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.pc = opdata;
			}
			break;

		case WT_TOMB_OP_EXEC_I: /* Execute microprogram Immediate */
			{
__exec_i:
				if(opdata<WTR_NUM_MICROPROGRAMS)
				{
					ctx->run = wtr_microprogtab[opdata](prof, ctx->frame.prog,
						ctx);
					goto handle_rr;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_EXEC_L: /* Execute microprogram LongImmediate */
			{
				READ_LONGIMM;
				goto __exec_i;
			}
			break;

		case WT_TOMB_OP_EXEC_S: /* Execute microprogram StackImmediate */
			{
				READ_STACKIMM;
				goto __exec_i;
			}
			break;

		case WT_TOMB_OP_DLE_I: /* Dynamic Linking Execution Immediate */
			{
				goto __dle_l;
			}
			break;

		case WT_TOMB_OP_DLE_L: /* Dynamic Linking Execution LongImmediate */
			{
				wt_pfn* pfn;
				READ_LONGIMM;
__dle_l:
				pfn = (wt_pfn*)WTR_PTR_FROM_VMADDR(ctx->frame.prog,
					ctx->frame.prog->base_address + opdata);
				ctx->run = wtr_processor_op_CFP(prof, ctx, pfn);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_DLE_S: /* Dynamic Linking Execution StackImmediate */
			{
				READ_STACKIMM;
				goto __dle_l;
			}
			break;

		case WT_TOMB_OP_CFP: /* Call Function Pointer */
			{
				wt_pfn pfn;
				CHECK_POP(ctx, sizeof(wt_pfn));
				pfn = *(wt_pfn*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_pfn);

				ctx->run = wtr_processor_op_CFP(prof, ctx, &pfn);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_MKFP_I: /* Make Function Pointer Immediate */
			{
				wt_pfn pfn;
__mkfp_i:
				/* check whether the assembled function pointer fits on the stack */
				CHECK_PUSH(ctx, sizeof(wt_pfn));

				pfn._pad = 0;
				pfn.ptr  = ctx->frame.prog;
				pfn.ext  = ctx->frame.prog->pc_base + opdata;

				WTR_CTX_DS_PUSH(ctx, wt_pfn, pfn);
			}
			break;

		case WT_TOMB_OP_MKFP_L: /* Make Function Pointer LongImmediate */
			{
				READ_LONGIMM;
				goto __mkfp_i;
			}
			break;

		case WT_TOMB_OP_MKFP_S: /* Make Function Pointer StackImmediate */
			{
				READ_STACKIMM;
				goto __mkfp_i;
			}
			break;

		case WT_TOMB_OP_TTD: /* TOS Top Delete */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					struct wtr_objentry ent = *wtr_dynstack_top(&ctx->tos);
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
						struct wtr_objentry* const top = wtr_dynstack_top(&ctx->rtos);
						wtr_dynstack_push(&ctx->tos, top);
						wtr_dynstack_pop(&ctx->rtos);
					}

					CHECK_PUSH(ctx, sizeof(wt_u32));
					WTR_CTX_DS_PUSH(ctx, wt_u32, ent.obj);

					ctx->run = wtr_processor_op_CFP(prof, ctx, &ent.pfn);
					goto handle_rr;
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_LDFP: /* Load Function Pointer */
			{
				wt_vmptr addr;
				CHECK_POP(ctx, sizeof(addr));
				addr = *(wt_vmptr*)ctx->dsp;
				if(WTR_CTX_DS_SPACE(ctx)>=sizeof(wt_pfn))
				{
					ctx->dsp -= (sizeof(wt_pfn) - sizeof(addr));
					memcpy(ctx->dsp, WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr), sizeof(wt_pfn));
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_STFP: /* Store Function Pointer */
			{
				wt_vmptr addr;
				CHECK_POP(ctx, sizeof(addr)+sizeof(wt_pfn));
				addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				memcpy(WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr), ctx->dsp, sizeof(wt_pfn));
				WTR_CTX_DS_POP(ctx, wt_pfn);
			}
			break;

		case WT_TOMB_OP_FPS_I: /* Function Pointer Size Immediate */
			{
__fps_i:
				opdata *= sizeof(wt_pfn);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_FPS_L: /* Function Pointer Size LongImmediate */
			{
				READ_LONGIMM;
				goto __fps_i;
			}
			break;

		case WT_TOMB_OP_STRN: /* New String */
			{
				wt_vmptr        vm_addr;
				struct wtr_str* str;
				char*           sdata = NULL;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);

				if(vm_addr)
				{
					sdata = (char*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, vm_addr);
				}

				if(NULL==(str=wtr_str_create(ctx->frame.prog, sdata, 0, WT_TRUE)))
				{
					goto__out_of_memory_exception();
				}

				vm_addr = WTR_OBJ_TO_VMADDR(ctx->frame.prog, str);
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, vm_addr);
			}
			break;

		case WT_TOMB_OP_STRC: /* Copy String */
			{
				wt_vmptr        vm_addr;
				struct wtr_str* str;
#if WTRUNTIME_DEBUG
				struct wtr_str* old;
#endif /* WTRUNTIME_DEBUG */

				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				str = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vm_addr);

#if WTRUNTIME_DEBUG
				old = str;
				(void)old;
#endif /* WTRUNTIME_DEBUG */

				ASSERT_OBJECT_EXISTENCE(str);

				if((str=wtr_str_create(prog, str->data, str->length, WT_FALSE))==NULL)
				{
					goto__out_of_memory_exception();
				}

				vm_addr = WTR_OBJ_TO_VMADDR(prog, str);
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, vm_addr);
			}
			break;

		case WT_TOMB_OP_STRL: /* Get String Length */
			{
				wt_vmptr vm_addr;
				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				WTR_CTX_DS_PUSH(ctx, wt_u32, ((struct wtr_str*)WTR_OBJ_FROM_VMADDR(ctx->frame.prog, vm_addr))->length);
			}
			break;

		case WT_TOMB_OP_STRD: /* Delete String */
			{
				wt_vmptr vm_addr;
				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);

				if(0!=vm_addr)
				{
					wtr_str_delete(prog, (struct wtr_str*)WTR_OBJ_FROM_VMADDR(prog, vm_addr));
				}
			}
			break;

		case WT_TOMB_OP_STRCR_I: /* String Character Reference Immediate */
			{
				goto __strcr_s;
			}
			break;

		case WT_TOMB_OP_STRCR_L: /* String Character Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __strcr_s;
			}
			break;

		case WT_TOMB_OP_STRCR_S: /* String Character Reference StackImmediate */
			{
				struct wtr_str* str;
				wt_vmptr        vmptr;
				READ_STACKIMM;
__strcr_s:
				CHECK_POP(ctx, sizeof(wt_vmptr));
				vmptr = *(wt_vmptr*)ctx->dsp;
				str = (struct wtr_str*)WTR_OBJ_FROM_VMADDR(ctx->frame.prog, vmptr);

				ASSERT_OBJECT_EXISTENCE(str);
				CHECK_OBJECT_MAGIC12_STR(str);

				if(opdata<str->length)
				{
					*(wt_vmptr*)ctx->dsp = WTR_PTR_TO_VMADDR(ctx->frame.prog, str->data+opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQN_I: /* New Sequence Skeleton Immediate */
			{
				struct wtr_seq* seq;
				wt_u32          cnt;
__seqn_i:
				CHECK_POP(ctx, sizeof(wt_u32));
				cnt = *(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);

				if(0<opdata)
				{
					struct wtr_program* const prog = ctx->frame.prog;

					if((seq=wtr_seq_create(prog, opdata, cnt))!=NULL)
					{
						WTR_CTX_DS_PUSH(ctx, wt_vmptr, WTR_OBJ_TO_VMADDR(prog, seq));
					}
					else
					{
						goto__out_of_memory_exception();
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQN_L: /* New Sequence Skeleton LongImmediate */
			{
				READ_LONGIMM;
				goto __seqn_i;
			}
			break;

		case WT_TOMB_OP_SEQN_S: /* New Sequence Skeleton StackImmediate */
			{
				READ_STACKIMM;
				goto __seqn_i;
			}
			break;

		case WT_TOMB_OP_SEQD: /* Delete Sequence Skeleton */
			{
				wt_vmptr vm_addr;
				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				wtr_seq_delete(prog, (struct wtr_seq*)WTR_OBJ_FROM_VMADDR(prog, vm_addr));
			}
			break;

		case WT_TOMB_OP_SEQIR_I: /* Sequence Item Reference Immediate */
			{
				goto __seqir_s;
			}
			break;

		case WT_TOMB_OP_SEQIR_L: /* Sequence Item Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __seqir_s;
			}
			break;

		case WT_TOMB_OP_SEQIR_S: /* Sequence Item Reference StackImmediate */
			{
				struct wtr_seq* seq;
				wt_vmptr        vmptr;
				READ_STACKIMM;
__seqir_s:
				CHECK_POP(ctx, sizeof(wt_vmptr));
				vmptr = *(wt_vmptr*)ctx->dsp;
				seq = (struct wtr_seq*)WTR_OBJ_FROM_VMADDR(ctx->frame.prog, vmptr);

				ASSERT_OBJECT_EXISTENCE(seq);
				CHECK_OBJECT_MAGIC12_SEQ(seq);

				if(opdata<seq->length)
				{
					*(wt_vmptr*)ctx->dsp = WTR_PTR_TO_VMADDR(ctx->frame.prog, seq->data+seq->elemsize*opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQL: /* Get Sequence Length */
			{
				struct wtr_seq* seq;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				seq = (struct wtr_seq*)WTR_OBJ_FROM_VMADDR(ctx->frame.prog, *(wt_vmptr*)ctx->dsp);

				ASSERT_OBJECT_EXISTENCE(seq);

				ctx->dsp -= (sizeof(wt_u32) - sizeof(wt_vmptr));

				*(wt_u32*)ctx->dsp = seq->length;
			}
			break;

		case WT_TOMB_OP_RECN_I: /* New Record Skeleton Immediate */
			{
				struct wtr_rec* rec;
__recn_i:
				if(NULL!=(rec=wtr_rec_create(ctx->frame.prog, opdata)))
				{
					WTR_CTX_DS_PUSH(ctx, wt_vmptr, WTR_OBJ_TO_VMADDR(ctx->frame.prog, rec));
				}
				else
				{
					goto__out_of_memory_exception();
				}
			}
			break;

		case WT_TOMB_OP_RECN_L: /* New Record Skeleton LongImmediate */
			{
				READ_LONGIMM;
				goto __recn_i;
			}
			break;

		case WT_TOMB_OP_RECN_S: /* New Record Skeleton StackImmediate */
			{
				READ_STACKIMM;
				goto __recn_i;
			}
			break;

		case WT_TOMB_OP_RECD: /* Delete Record Skeleton */
			{
				wt_vmptr vm_addr;
				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				vm_addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				wtr_rec_delete(prog, (struct wtr_rec*)WTR_OBJ_FROM_VMADDR(prog, vm_addr));
			}
			break;

		case WT_TOMB_OP_RECFR_I: /* Record Field Reference Immediate */
			{
				struct wtr_rec* rec;
				wt_vmptr        vmptr;
__recfr_i:
				CHECK_POP(ctx, sizeof(wt_vmptr));
				vmptr = *(wt_vmptr*)ctx->dsp;
				rec = (struct wtr_rec*)WTR_OBJ_FROM_VMADDR(ctx->frame.prog, vmptr);

				ASSERT_OBJECT_EXISTENCE(rec);
				CHECK_OBJECT_MAGIC12_REC(rec);

				if(opdata<rec->size)
				{
					*(wt_vmptr*)ctx->dsp = WTR_PTR_TO_VMADDR(ctx->frame.prog, rec->data+opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_RECFR_L: /* Record Field Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __recfr_i;
			}
			break;

		case WT_TOMB_OP_RECFR_S: /* Record Field Reference StackImmediate */
			{
				READ_STACKIMM;
				goto __recfr_i;
			}
			break;

		case WT_TOMB_OP_ERAISE: /* Raise Exception */
			{
				ctx->last_exception = *(wt_exception*)ctx->dsp;
				goto unwind_frames_until_ehandler;
			}
			break;

		case WT_TOMB_OP_EXHAN_I: /* set Exception Handler Immediate */
			{
				ctx->frame.ehandler = ctx->frame.pc - 1 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EXHAN_L: /* set Exception Handler LongImmediate */
			{
				READ_LONGIMM;
				ctx->frame.ehandler = ctx->frame.pc - 2 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EXHAN_S: /* set Exception Handler StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.ehandler = ctx->frame.pc - 1 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EHCL: /* Clear Exception Handler */
			{
				ctx->frame.ehandler = WTR_NULLPC;
				ctx->frame.dsp_mark = NULL;
			}
			break;

		_OP_I_S(ADD, wt_u32, NOTHING)
		_OP_I_S(SUB, wt_u32, NOTHING)
		_OP_I_S(MUL, wt_u32, NOTHING)
		_OP_I_S(DIV, wt_u32, DIV_NONZERO)
		_OP_I_S(SHL, wt_u32, NOTHING)
		_OP_I_S(SHR, wt_u32, NOTHING)
		_OP_I_S(SAR, wt_i32, NOTHING)
		_OP_I_S(ROR, wt_u32, NOTHING)
		_OP_I_S(SMUL, wt_i32, NOTHING)
		_OP_I_S(SDIV, wt_i32, DIV_NONZERO)
		_OP_(MOD, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(SMOD, wt_i32, wt_i32, DIV_NONZERO)
		_OP_I_L(LSOR, wt_u32, NOTHING)
		_OP_I_L(RSB, wt_u32, NOTHING)
		_OP_(ADDL, wt_u64, wt_u64, NOTHING)
		_OP_(SUBL, wt_u64, wt_u64, NOTHING)
		_OP_(MULL, wt_u64, wt_u64, NOTHING)
		_OP_(DIVL, wt_u64, wt_u64, DIV_NONZERO)
		_OP_(SHLL, wt_u64, wt_u64, NOTHING)
		_OP_(SHRL, wt_u64, wt_u64, NOTHING)
		_OP_(SARL, wt_u64, wt_u64, NOTHING)
		_OP_(RORL, wt_u64, wt_u64, NOTHING)
		_OP_(XOR, wt_u32, wt_u32, NOTHING)
		_OP_(BAND, wt_u32, wt_u32, NOTHING)
		_OP_(BOR, wt_u32, wt_u32, NOTHING)

		case WT_TOMB_OP_BNOT:
			{
				CHECK_POP(ctx, sizeof(wt_u32));
				*(wt_u32*)ctx->dsp = ~(*(wt_u32*)ctx->dsp);
			}
			break;

		_OP_(XORL, wt_u64, wt_u64, NOTHING)
		_OP_(BANDL, wt_u64, wt_u64, NOTHING)
		_OP_(BORL, wt_u64, wt_u64, NOTHING)
		_OP_(MODL, wt_u64, wt_u64, DIV_NONZERO)
		_OP_(SMODL, wt_i64, wt_i64, DIV_NONZERO)

		case WT_TOMB_OP_BNOTL:
			{
				CHECK_POP(ctx, sizeof(wt_u64));
				*(wt_u64*)ctx->dsp = ~(*(wt_u64*)ctx->dsp);
			}
			break;

		_OP_I_L(LSORL, wt_u64, NOTHING)
		_OP_(SMULL, wt_i64, wt_i64, NOTHING)
		_OP_(SDIVL, wt_i64, wt_i64, DIV_NONZERO)

		case WT_TOMB_OP_NEG:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_i32*)ctx->dsp = -(*(wt_i32*)ctx->dsp);
			}
			break;

		case WT_TOMB_OP_NEGL:
			{
				CHECK_POP(ctx, sizeof(wt_i64));
				*(wt_i64*)ctx->dsp = -(*(wt_i64*)ctx->dsp);
			}
			break;

#if WTRUNTIME_ENABLE_FLOAT32
		_OP_(FADD, wt_f32, wt_f32, NOTHING)
		_OP_(FSUB, wt_f32, wt_f32, NOTHING)
		_OP_(FMUL, wt_f32, wt_f32, NOTHING)
		_OP_(FDIV, wt_f32, wt_f32, DIV_NONZERO)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
		_OP_(FADDL, wt_f64, wt_f64, NOTHING)
		_OP_(FSUBL, wt_f64, wt_f64, NOTHING)
		_OP_(FMULL, wt_f64, wt_f64, NOTHING)
		_OP_(FDIVL, wt_f64, wt_f64, DIV_NONZERO)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FNEG:
			{
				CHECK_POP(ctx, sizeof(wt_f32));
				*(wt_f32*)ctx->dsp = -(*(wt_f32*)ctx->dsp);
			}
			break;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64
		case WT_TOMB_OP_FNEGL:
			{
				CHECK_POP(ctx, sizeof(wt_f64));
				*(wt_f64*)ctx->dsp = -(*(wt_f64*)ctx->dsp);
			}
			break;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32
		_OP_(FMOD, wt_f32, wt_f32, DIV_NONZERO)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
		_OP_(FMODL, wt_f64, wt_f64, DIV_NONZERO)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FABS:
			{
				CHECK_POP(ctx, sizeof(wt_f32));
				if((*(wt_f32*)ctx->dsp) < 0.f)
				{
					*(wt_f32*)ctx->dsp = -(*(wt_f32*)ctx->dsp);
				}
			}
			break;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64
		case WT_TOMB_OP_FABSL:
			{
				CHECK_POP(ctx, sizeof(wt_f64));
				if((*(wt_f64*)ctx->dsp) < 0.0)
				{
					*(wt_f64*)ctx->dsp = -(*(wt_f64*)ctx->dsp);
				}
			}
			break;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

		_OP_cmp(CMP, wt_u32)
		_OP_cmp(SCMP, wt_i32)
		_OP_cmp(CMPL, wt_u64)
		_OP_cmp(SCMPL, wt_i64)
#if WTRUNTIME_ENABLE_FLOAT32
		_OP_cmp(FCMP, wt_f32)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
		_OP_cmp(FCMPL, wt_f64)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
		_OP_cmp(SCMPB, wt_u32)
		_OP_cmp(SCMPH, wt_u32)

		case WT_TOMB_OP_QLT:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) < 0;
			}
			break;

		case WT_TOMB_OP_QLE:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) <= 0;
			}
			break;

		case WT_TOMB_OP_QGT:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) > 0;
			}
			break;

		case WT_TOMB_OP_QGE:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) >= 0;
			}
			break;

		case WT_TOMB_OP_QEQ:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) == 0;
			}
			break;

		case WT_TOMB_OP_QNE:
			{
				CHECK_POP(ctx, sizeof(wt_i32));
				*(wt_bool*)ctx->dsp = (*(wt_i32*)ctx->dsp) != 0;
			}
			break;

		case WT_TOMB_OP_OSDR:
			{
				if(WTRCF_CSOS & ctx->frame.flags)
				{
					if(ctx->frame.csos_sp)
					{
						--ctx->frame.csos_sp;
					}
					else
					{
						goto stack_underflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_MSKBTM:
			{
				CHECK_POP(ctx, sizeof(wt_u32));

				if(32>opdata)
				{
					*(wt_u32*)ctx->dsp &= (((wt_u32)1) << opdata) - 1;
				}
				else
				if(32==opdata)
				{
					/* nop */
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_LNOT:
			{
				CHECK_POP(ctx, sizeof(wt_bool));
				*(wt_bool*)ctx->dsp = !(*(wt_bool*)ctx->dsp);
			}
			break;

		_OP_(LAND, wt_bool, wt_bool, NOTHING)
		_OP_(LOR, wt_bool, wt_bool, NOTHING)
		_OP_(LXOR, wt_bool, wt_bool, NOTHING)

		case WT_TOMB_OP_LDB:
			{
				wt_vmptr addr;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				addr = *(wt_vmptr*)ctx->dsp;
				ctx->dsp -= (sizeof(wt_u32) - sizeof(wt_vmptr));
				*(wt_u32*)ctx->dsp = (wt_u32)(*(wt_u8*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr));
			}
			break;

		case WT_TOMB_OP_STB:
			{
				wt_vmptr addr;
				wt_u32   word;

				CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));
				addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				word = *(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);
				*(wt_u8*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr) = (wt_u8)word;
			}
			break;

		case WT_TOMB_OP_LDH:
			{
				wt_vmptr addr;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				addr = *(wt_vmptr*)ctx->dsp;
				ctx->dsp -= (sizeof(wt_u32) - sizeof(wt_vmptr));
				*(wt_u32*)ctx->dsp = (wt_u32)(*(wt_u16*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr));
			}
			break;

		case WT_TOMB_OP_STH:
			{
				wt_vmptr addr;
				wt_u32   word;

				CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));
				addr = *(wt_vmptr*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_vmptr);
				word = *(wt_u32*)ctx->dsp;
				WTR_CTX_DS_POP(ctx, wt_u32);
				*(wt_u16*)WTR_PTR_FROM_VMADDR(ctx->frame.prog, addr) = (wt_u16)word;
			}
			break;

		_OP_I_S(ADDB, wt_u32, NOTHING)
		_OP_I_S(SUBB, wt_u32, NOTHING)
		_OP_(MULB, wt_u32, wt_u32, NOTHING)
		_OP_(SMULB, wt_u32, wt_u32, NOTHING)
		_OP_(DIVB, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(SDIVB, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(MODB, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(SMODB, wt_u32, wt_u32, DIV_NONZERO)
		_OP_I_S(SHLB, wt_u32, NOTHING)
		_OP_(RORB, wt_u32, wt_u32, NOTHING)

		case WT_TOMB_OP_BNOTB:
			{
				CHECK_POP(ctx, sizeof(wt_u32));
				*(wt_u32*)ctx->dsp = ~((wt_u8)(*(wt_u32*)ctx->dsp));
			}
			break;

		_OP_I_S(ADDH, wt_u32, NOTHING)
		_OP_I_S(SUBH, wt_u32, NOTHING)
		_OP_(MULH, wt_u32, wt_u32, NOTHING)
		_OP_(SMULH, wt_u32, wt_u32, NOTHING)
		_OP_(DIVH, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(SDIVH, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(MODH, wt_u32, wt_u32, DIV_NONZERO)
		_OP_(SMODH, wt_u32, wt_u32, DIV_NONZERO)
		_OP_I_S(SHLH, wt_u32, NOTHING)
		_OP_(RORH, wt_u32, wt_u32, NOTHING)

		case WT_TOMB_OP_BNOTH:
			{
				CHECK_POP(ctx, sizeof(wt_u32));
				*(wt_u32*)ctx->dsp = ~((wt_u16)(*(wt_u32*)ctx->dsp));
			}
			break;

		case WT_TOMB_OP_SGNXD:
			{
				CHECK_POP(ctx, sizeof(wt_u32));

				if(32>opdata && 0!=opdata)
				{
					const wt_u32 signbit = ((wt_u32)1) << (opdata - 1);

					if(signbit & *(wt_u32*)ctx->dsp)
					{
						*(wt_u32*)ctx->dsp |= ~(signbit - 1);
					}
				}
				else
				if(32==opdata)
				{
					/* nop */
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		default:
			{
				/* undefined instruction */

				exc = UndefinedInstructionException;
				goto raise_exception;

handle_rr:
				switch(ctx->run)
				{
				case WTRR_RETURNED:
				case WTRR_EXCEPTION:
					return ctx->run;

				case WTRR_RAISE:
					exc = ctx->last_exception;
					goto raise_exception;

				default:
					break;
				}
				break;

io_exception:
				exc = IOException;
				goto raise_exception;

access_violation_exception:
				exc = MemoryAccessViolationException;
				goto raise_exception;

bad_program_access_exception:
				exc = BadProgramAccessException;
				goto raise_exception;

stack_overflow_exception:
stack_underflow_exception:
				exc = StackOverflowException;
				goto raise_exception;

division_by_zero_exception:
				exc = DivisionByZeroException;
				goto raise_exception;

index_out_of_range_exception:
				exc = IndexOutOfRangeException;
				goto raise_exception;

out_of_memory_exception:
				exc = OutOfMemoryException;
				goto raise_exception;

raise_exception:
				WTR_CTX_DS_PUSH(ctx, wt_exception, exc); /* if the stack is full, there is still a spare space for the exception */
unwind_frames_until_ehandler:
				ctx->run = wtr_processor_unwind_frames(prof, ctx);
				goto handle_rr;

#ifdef WTRUNTIME_DEBUG
				opcode = opcode;
#endif /* WTRUNTIME_DEBUG */
			}
			break;
		}
	}

	return ctx->run;
}
